# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/10

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The account option for the journal entry management.

"""
from accounting.models import Account


class AccountOption:
    """An account option."""

    def __init__(self, account: Account):
        """Constructs an account option.

        :param account: The account.
        """
        self.id: str = account.id
        """The account ID."""
        self.code: str = account.code
        """The account code."""
        self.title: str = account.title
        """The account title."""
        self.query_values: list[str] = account.query_values
        """The values to be queried."""
        self.__str: str = str(account)
        """The string representation of the account option."""
        self.is_in_use: bool = False
        """True if this account is in use, or False otherwise."""
        self.is_need_offset: bool = account.is_need_offset
        """True if this account needs offset, or False otherwise."""

    def __str__(self) -> str:
        """Returns the string representation of the account option.

        :return: The string representation of the account option.
        """
        return self.__str



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/1

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The path converters for the base account management.

"""
from flask import abort
from werkzeug.routing import BaseConverter

from accounting import db
from accounting.models import BaseAccount


class BaseAccountConverter(BaseConverter):
    """The account converter to convert the account code and to the
    corresponding base account in the routes."""

    def to_python(self, value: str) -> BaseAccount:
        """Converts an account code to a base account.

        :param value: The account code.
        :return: The corresponding base account.
        """
        account: BaseAccount | None = db.session.get(BaseAccount, value)
        if account is None:
            abort(404)
        return account

    def to_url(self, value: BaseAccount) -> str:
        """Converts a base account to its code.

        :param value: The base account.
        :return: The code.
        """
        return value.code



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/18

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The journal entry forms for the journal entry management.

"""
import datetime as dt
from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Type

import sqlalchemy as sa
from flask_babel import LazyString
from flask_wtf import FlaskForm
from wtforms import DateField, FieldList, FormField, TextAreaField, \
    BooleanField
from wtforms.validators import DataRequired, ValidationError

from accounting import db
from accounting.journal_entry.utils.account_option import AccountOption
from accounting.journal_entry.utils.description_editor import DescriptionEditor
from accounting.journal_entry.utils.original_line_items import \
    get_selectable_original_line_items
from accounting.locale import lazy_gettext
from accounting.models import JournalEntry, Account, JournalEntryLineItem, \
    JournalEntryCurrency
from accounting.utils.random_id import new_id
from accounting.utils.strip_text import strip_multiline_text
from accounting.utils.user import get_current_user_pk
from .currency import CurrencyForm, CashReceiptCurrencyForm, \
    CashDisbursementCurrencyForm, TransferCurrencyForm
from .line_item import LineItemForm, DebitLineItemForm, CreditLineItemForm
from .reorder import sort_journal_entries_in

DATE_REQUIRED: DataRequired = DataRequired(
    lazy_gettext("Please fill in the date."))
"""The validator to check if the date is empty."""


class NotBeforeOriginalLineItems:
    """The validator to check if the date is not before the
    original line items."""

    def __call__(self, form: FlaskForm, field: DateField) -> None:
        assert isinstance(form, JournalEntryForm)
        if field.data is None:
            return
        min_date: dt.date | None = form.min_date
        if min_date is None:
            return
        if field.data < min_date:
            raise ValidationError(lazy_gettext(
                "The date cannot be earlier than the original line items."))


class NotAfterOffsetItems:
    """The validator to check if the date is not after the offset items."""

    def __call__(self, form: FlaskForm, field: DateField) -> None:
        assert isinstance(form, JournalEntryForm)
        if field.data is None:
            return
        max_date: dt.date | None = form.max_date
        if max_date is None:
            return
        if field.data > max_date:
            raise ValidationError(lazy_gettext(
                "The date cannot be later than the offset items."))


class NeedSomeCurrencies:
    """The validator to check if there is any currency sub-form."""

    def __call__(self, form: FlaskForm, field: FieldList) -> None:
        if len(field) == 0:
            raise ValidationError(lazy_gettext("Please add some currencies."))


class CannotDeleteOriginalLineItemsWithOffset:
    """The validator to check the original line items with offset."""

    def __call__(self, form: FlaskForm, field: FieldList) -> None:
        assert isinstance(form, JournalEntryForm)
        if form.obj is None:
            return
        existing_matched_original_line_item_id: set[int] \
            = {x.id for x in form.obj.line_items if len(x.offsets) > 0}
        line_item_id_in_form: set[int] \
            = {x.id.data for x in form.line_items if x.id.data is not None}
        for line_item_id in existing_matched_original_line_item_id:
            if line_item_id not in line_item_id_in_form:
                raise ValidationError(lazy_gettext(
                    "Line items with offset cannot be deleted."))


class JournalEntryForm(FlaskForm):
    """The base form to create or edit a journal entry."""
    date = DateField()
    """The date."""
    currencies = FieldList(FormField(CurrencyForm))
    """The line items categorized by their currencies."""
    note = TextAreaField()
    """The note."""

    def __init__(self, *args, **kwargs):
        """Constructs a base journal entry form.

        :param args: The arguments.
        :param kwargs: The keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self.is_modified: bool = False
        """Whether the journal entry is modified during populate_obj()."""
        self.collector: Type[LineItemCollector] = LineItemCollector
        """The line item collector.  The default is the base abstract
        collector only to provide the correct type.  The subclass forms should
        provide their own collectors."""
        self.obj: JournalEntry | None = kwargs.get("obj")
        """The journal entry, when editing an existing one."""
        self._is_need_payable: bool = False
        """Whether we need the payable original line items."""
        self._is_need_receivable: bool = False
        """Whether we need the receivable original line items."""
        self.__original_line_item_options: list[JournalEntryLineItem] | None \
            = None
        """The options of the original line items."""
        self.__net_balance_exceeded: dict[int, LazyString] | None = None
        """The original line items whose net balances were exceeded by the
        amounts in the line item sub-forms."""
        for line_item in self.line_items:
            line_item.journal_entry_form = self

    def populate_obj(self, obj: JournalEntry) -> None:
        """Populates the form data into a journal entry object.

        :param obj: The journal entry object.
        :return: None.
        """
        is_new: bool = obj.id is None
        if is_new:
            obj.id = new_id(JournalEntry)
        self.__set_date(obj, self.date.data)
        obj.note = self.note.data

        collector_cls: Type[LineItemCollector] = self.collector
        collector: collector_cls = collector_cls(self, obj)
        collector.collect()

        to_delete: set[int] = {x.id for x in obj.line_items
                               if x.id not in collector.to_keep}
        if len(to_delete) > 0:
            JournalEntryLineItem.query\
                .filter(JournalEntryLineItem.id.in_(to_delete)).delete()
            self.is_modified = True

        if is_new or db.session.is_modified(obj):
            self.is_modified = True

        if is_new:
            current_user_pk: int = get_current_user_pk()
            obj.created_by_id = current_user_pk
            obj.updated_by_id = current_user_pk

    @property
    def line_items(self) -> list[LineItemForm]:
        """Collects and returns the line item sub-forms.

        :return: The line item sub-forms.
        """
        line_items: list[LineItemForm] = []
        for currency in self.currencies:
            line_items.extend(currency.line_items)
        return line_items

    def __set_date(self, obj: JournalEntry, new_date: dt.date) -> None:
        """Sets the journal entry date and number.

        :param obj: The journal entry object.
        :param new_date: The new date.
        :return: None.
        """
        if obj.date is None or obj.date != new_date:
            if obj.date is not None:
                sort_journal_entries_in(obj.date, obj.id)
            if self.max_date is not None and new_date == self.max_date:
                db_min_no: int | None = db.session.scalar(
                    sa.select(sa.func.min(JournalEntry.no))
                    .filter(JournalEntry.date == new_date))
                if db_min_no is None:
                    obj.date = new_date
                    obj.no = 1
                else:
                    obj.date = new_date
                    obj.no = db_min_no - 1
                    sort_journal_entries_in(new_date)
            else:
                sort_journal_entries_in(new_date, obj.id)
                count: int = JournalEntry.query\
                    .filter(JournalEntry.date == new_date).count()
                obj.date = new_date
                obj.no = count + 1

    @property
    def debit_account_options(self) -> list[AccountOption]:
        """The selectable debit accounts.

        :return: The selectable debit accounts.
        """
        accounts: list[AccountOption] \
            = [AccountOption(x) for x in Account.selectable_debit()
               if not (x.code[0] == "2" and x.is_need_offset)]
        in_use: set[int] = set(db.session.scalars(
            sa.select(JournalEntryLineItem.account_id)
            .filter(JournalEntryLineItem.is_debit)
            .group_by(JournalEntryLineItem.account_id)).all())
        for account in accounts:
            account.is_in_use = account.id in in_use
        return accounts

    @property
    def credit_account_options(self) -> list[AccountOption]:
        """The selectable credit accounts.

        :return: The selectable credit accounts.
        """
        accounts: list[AccountOption] \
            = [AccountOption(x) for x in Account.selectable_credit()
               if not (x.code[0] == "1" and x.is_need_offset)]
        in_use: set[int] = set(db.session.scalars(
            sa.select(JournalEntryLineItem.account_id)
            .filter(sa.not_(JournalEntryLineItem.is_debit))
            .group_by(JournalEntryLineItem.account_id)).all())
        for account in accounts:
            account.is_in_use = account.id in in_use
        return accounts

    @property
    def currencies_errors(self) -> list[str | LazyString]:
        """Returns the currency errors, without the errors in their sub-forms.

        :return: The currency errors, without the errors in their sub-forms.
        """
        return [x for x in self.currencies.errors
                if isinstance(x, str) or isinstance(x, LazyString)]

    @property
    def description_editor(self) -> DescriptionEditor:
        """Returns the description editor.

        :return: The description editor.
        """
        return DescriptionEditor()

    @property
    def original_line_item_options(self) -> list[JournalEntryLineItem]:
        """Returns the selectable original line items.

        :return: The selectable original line items.
        """
        if self.__original_line_item_options is None:
            self.__original_line_item_options \
                = get_selectable_original_line_items(
                    {x.id.data for x in self.line_items
                     if x.id.data is not None},
                    self._is_need_payable, self._is_need_receivable)
        return self.__original_line_item_options

    @property
    def min_date(self) -> dt.date | None:
        """Returns the minimal available date.

        :return: The minimal available date.
        """
        original_line_item_id: set[int] \
            = {x.original_line_item_id.data for x in self.line_items
               if x.original_line_item_id.data is not None}
        if len(original_line_item_id) == 0:
            return None
        select: sa.Select = sa.select(sa.func.max(JournalEntry.date))\
            .join(JournalEntryLineItem)\
            .filter(JournalEntryLineItem.id.in_(original_line_item_id))
        return db.session.scalar(select)

    @property
    def max_date(self) -> dt.date | None:
        """Returns the maximum available date.

        :return: The maximum available date.
        """
        line_item_id: set[int] = {x.id.data for x in self.line_items
                                  if x.id.data is not None}
        select: sa.Select = sa.select(sa.func.min(JournalEntry.date))\
            .join(JournalEntryLineItem)\
            .filter(JournalEntryLineItem.original_line_item_id
                    .in_(line_item_id))
        return db.session.scalar(select)


T = TypeVar("T", bound=JournalEntryForm)
"""A journal entry form variant."""


class LineItemCollector(Generic[T], ABC):
    """The line item collector."""

    def __init__(self, form: T, obj: JournalEntry):
        """Constructs the line item collector.

        :param form: The journal entry form.
        :param obj: The journal entry.
        """
        self.form: T = form
        """The journal entry form."""
        self.__obj: JournalEntry = obj
        """The journal entry object."""
        self.__line_items: list[JournalEntryLineItem] = list(obj.line_items)
        """The existing line items."""
        self.__line_items_by_id: dict[int, JournalEntryLineItem] \
            = {x.id: x for x in self.__line_items}
        """A dictionary from the line item ID to their line items."""
        self.__no_by_id: dict[int, int] \
            = {x.id: x.no for x in self.__line_items}
        """A dictionary from the line item number to their line items."""
        self.__currencies: list[JournalEntryCurrency] = obj.currencies
        """The currencies in the journal entry."""
        self._debit_no: int = 1
        """The number index for the debit line items."""
        self._credit_no: int = 1
        """The number index for the credit line items."""
        self.to_keep: set[int] = set()
        """The ID of the existing line items to keep."""

    @abstractmethod
    def collect(self) -> set[int]:
        """Collects the line items.

        :return: The ID of the line items to keep.
        """

    def _add_line_item(self, form: LineItemForm, currency_code: str, no: int) \
            -> None:
        """Composes a line item from the form.

        :param form: The line item form.
        :param currency_code: The code of the currency.
        :param no: The number of the line item.
        :return: None.
        """
        line_item: JournalEntryLineItem | None \
            = self.__line_items_by_id.get(form.id.data)
        if line_item is not None:
            line_item.currency_code = currency_code
            form.populate_obj(line_item)
            line_item.no = no
            if db.session.is_modified(line_item):
                self.form.is_modified = True
        else:
            line_item = JournalEntryLineItem()
            line_item.currency_code = currency_code
            form.populate_obj(line_item)
            line_item.no = no
            self.__obj.line_items.append(line_item)
            self.form.is_modified = True
        self.to_keep.add(line_item.id)

    def _make_cash_line_item(self, forms: list[LineItemForm], is_debit: bool,
                             currency_code: str, no: int) -> None:
        """Composes the cash line item at the other debit or credit of the
        cash journal entry.

        :param forms: The line item forms in the same currency.
        :param is_debit: True for a cash receipt journal entry, or False for a
            cash disbursement journal entry.
        :param currency_code: The code of the currency.
        :param no: The number of the line item.
        :return: None.
        """
        candidates: list[JournalEntryLineItem] \
            = [x for x in self.__line_items
               if x.is_debit == is_debit and x.currency_code == currency_code]
        line_item: JournalEntryLineItem
        if len(candidates) > 0:
            candidates.sort(key=lambda x: x.no)
            line_item = candidates[0]
            line_item.account_id = Account.cash().id
            line_item.description = None
            line_item.amount = sum([x.amount.data for x in forms])
            line_item.no = no
            if db.session.is_modified(line_item):
                self.form.is_modified = True
        else:
            line_item = JournalEntryLineItem()
            line_item.id = new_id(JournalEntryLineItem)
            line_item.is_debit = is_debit
            line_item.currency_code = currency_code
            line_item.account_id = Account.cash().id
            line_item.description = None
            line_item.amount = sum([x.amount.data for x in forms])
            line_item.no = no
            self.__obj.line_items.append(line_item)
            self.form.is_modified = True
        self.to_keep.add(line_item.id)

    def _sort_line_item_forms(self, forms: list[LineItemForm]) -> None:
        """Sorts the line item sub-forms.

        :param forms: The line item sub-forms.
        :return: None.
        """
        missing_no: int = 100 if len(self.__no_by_id) == 0 \
            else max(self.__no_by_id.values()) + 100
        ord_by_form: dict[LineItemForm, int] \
            = {forms[i]: i for i in range(len(forms))}
        recv_no: set[int] = {x.no.data for x in forms if x.no.data is not None}
        missing_recv_no: int = 100 if len(recv_no) == 0 else max(recv_no) + 100
        forms.sort(key=lambda x: (x.no.data or missing_recv_no,
                                  missing_no if x.id.data is None else
                                  self.__no_by_id.get(x.id.data, missing_no),
                                  ord_by_form.get(x)))

    def _sort_currency_forms(self, forms: list[CurrencyForm]) -> None:
        """Sorts the currency forms.

        :param forms: The currency forms.
        :return: None.
        """
        missing_no: int = len(self.__currencies) + 100
        no_by_code: dict[str, int] = {self.__currencies[i].code: i
                                      for i in range(len(self.__currencies))}
        ord_by_form: dict[CurrencyForm, int] \
            = {forms[i]: i for i in range(len(forms))}
        recv_no: set[int] = {x.no.data for x in forms if x.no.data is not None}
        missing_recv_no: int = 100 if len(recv_no) == 0 else max(recv_no) + 100
        forms.sort(key=lambda x: (x.no.data or missing_recv_no,
                                  no_by_code.get(x.code.data, missing_no),
                                  ord_by_form.get(x)))


class CashReceiptJournalEntryForm(JournalEntryForm):
    """The form to create or edit a cash receipt journal entry."""
    date = DateField(
        validators=[DATE_REQUIRED,
                    NotBeforeOriginalLineItems(),
                    NotAfterOffsetItems()])
    """The date."""
    currencies = FieldList(FormField(CashReceiptCurrencyForm), name="currency",
                           validators=[NeedSomeCurrencies()])
    """The line items categorized by their currencies."""
    note = TextAreaField(filters=[strip_multiline_text])
    """The note."""
    whole_form = BooleanField(
        validators=[CannotDeleteOriginalLineItemsWithOffset()])
    """The pseudo field for the whole form validators."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_need_receivable = True

        class Collector(LineItemCollector[CashReceiptJournalEntryForm]):
            """The line item collector for the cash receipt journal entries."""

            def collect(self) -> None:
                currencies: list[CashReceiptCurrencyForm] \
                    = [x.form for x in self.form.currencies]
                self._sort_currency_forms(currencies)
                for currency in currencies:
                    # The debit cash line item
                    self._make_cash_line_item(list(currency.credit), True,
                                              currency.code.data,
                                              self._debit_no)
                    self._debit_no = self._debit_no + 1

                    # The credit forms
                    credit_forms: list[CreditLineItemForm] \
                        = [x.form for x in currency.credit]
                    self._sort_line_item_forms(credit_forms)
                    for credit_form in credit_forms:
                        self._add_line_item(credit_form, currency.code.data,
                                            self._credit_no)
                        self._credit_no = self._credit_no + 1

        self.collector = Collector


class CashDisbursementJournalEntryForm(JournalEntryForm):
    """The form to create or edit a cash disbursement journal entry."""
    date = DateField(
        validators=[DATE_REQUIRED,
                    NotBeforeOriginalLineItems(),
                    NotAfterOffsetItems()])
    """The date."""
    currencies = FieldList(FormField(CashDisbursementCurrencyForm),
                           name="currency",
                           validators=[NeedSomeCurrencies()])
    """The line items categorized by their currencies."""
    note = TextAreaField(filters=[strip_multiline_text])
    """The note."""
    whole_form = BooleanField(
        validators=[CannotDeleteOriginalLineItemsWithOffset()])
    """The pseudo field for the whole form validators."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_need_payable = True

        class Collector(LineItemCollector[CashDisbursementJournalEntryForm]):
            """The line item collector for the cash disbursement journal
            entries."""

            def collect(self) -> None:
                currencies: list[CashDisbursementCurrencyForm] \
                    = [x.form for x in self.form.currencies]
                self._sort_currency_forms(currencies)
                for currency in currencies:
                    # The debit forms
                    debit_forms: list[DebitLineItemForm] \
                        = [x.form for x in currency.debit]
                    self._sort_line_item_forms(debit_forms)
                    for debit_form in debit_forms:
                        self._add_line_item(debit_form, currency.code.data,
                                            self._debit_no)
                        self._debit_no = self._debit_no + 1

                    # The credit forms
                    self._make_cash_line_item(list(currency.debit), False,
                                              currency.code.data,
                                              self._credit_no)
                    self._credit_no = self._credit_no + 1

        self.collector = Collector


class TransferJournalEntryForm(JournalEntryForm):
    """The form to create or edit a transfer journal entry."""
    date = DateField(
        validators=[DATE_REQUIRED,
                    NotBeforeOriginalLineItems(),
                    NotAfterOffsetItems()])
    """The date."""
    currencies = FieldList(FormField(TransferCurrencyForm), name="currency",
                           validators=[NeedSomeCurrencies()])
    """The line items categorized by their currencies."""
    note = TextAreaField(filters=[strip_multiline_text])
    """The note."""
    whole_form = BooleanField(
        validators=[CannotDeleteOriginalLineItemsWithOffset()])
    """The pseudo field for the whole form validators."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_need_payable = True
        self._is_need_receivable = True

        class Collector(LineItemCollector[TransferJournalEntryForm]):
            """The line item collector for the transfer journal entries."""

            def collect(self) -> None:
                currencies: list[TransferCurrencyForm] \
                    = [x.form for x in self.form.currencies]
                self._sort_currency_forms(currencies)
                for currency in currencies:
                    # The debit forms
                    debit_forms: list[DebitLineItemForm] \
                        = [x.form for x in currency.debit]
                    self._sort_line_item_forms(debit_forms)
                    for debit_form in debit_forms:
                        self._add_line_item(debit_form, currency.code.data,
                                            self._debit_no)
                        self._debit_no = self._debit_no + 1

                    # The credit forms
                    credit_forms: list[CreditLineItemForm] \
                        = [x.form for x in currency.credit]
                    self._sort_line_item_forms(credit_forms)
                    for credit_form in credit_forms:
                        self._add_line_item(credit_form, currency.code.data,
                                            self._credit_no)
                        self._credit_no = self._credit_no + 1

        self.collector = Collector



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/4

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The period specification parser.

"""
import calendar
import datetime as dt
import re
from collections.abc import Callable
from typing import Type

from .period import Period
from .shortcuts import ThisMonth, LastMonth, SinceLastMonth, ThisYear, \
    LastYear, Today, Yesterday, AllTime

DATE_SPEC_RE: str = r"(\d{4})(?:-(\d{2})(?:-(\d{2}))?)?"
"""The regular expression of a date specification."""


def get_period(spec: str | None = None) -> Period:
    """Returns a period instance.

    :param spec: The period specification, or omit for the default.
    :return: The period instance.
    :raise ValueError: When the period specification is invalid.
    """
    if spec is None:
        return ThisMonth()
    named_periods: dict[str, Type[Callable[[], Period]]] = {
        "this-month": lambda: ThisMonth(),
        "last-month": lambda: LastMonth(),
        "since-last-month": lambda: SinceLastMonth(),
        "this-year": lambda: ThisYear(),
        "last-year": lambda: LastYear(),
        "today": lambda: Today(),
        "yesterday": lambda: Yesterday(),
        "all-time": lambda: AllTime(),
    }
    if spec in named_periods:
        return named_periods[spec]()
    start, end = __parse_spec(spec)
    if start is not None and end is not None and start > end:
        raise ValueError
    return Period(start, end)


def __parse_spec(text: str) -> tuple[dt.date | None, dt.date | None]:
    """Parses the period specification.

    :param text: The period specification.
    :return: The start and end day of the period.  The start and end day
        may be None.
    :raise ValueError: When the date is invalid.
    """
    if text == "-":
        return None, None
    m = re.match(f"^{DATE_SPEC_RE}$", text)
    if m is not None:
        return __get_start(m[1], m[2], m[3]), \
               __get_end(m[1], m[2], m[3])
    m = re.match(f"^{DATE_SPEC_RE}-$", text)
    if m is not None:
        return __get_start(m[1], m[2], m[3]), None
    m = re.match(f"-{DATE_SPEC_RE}$", text)
    if m is not None:
        return None, __get_end(m[1], m[2], m[3])
    m = re.match(f"^{DATE_SPEC_RE}-{DATE_SPEC_RE}$", text)
    if m is not None:
        return __get_start(m[1], m[2], m[3]), \
               __get_end(m[4], m[5], m[6])
    raise ValueError


def __get_start(year: str, month: str | None, day: str | None) -> dt.date:
    """Returns the start of the period from the date representation.

    :param year: The year.
    :param month: The month, if any.
    :param day: The day, if any.
    :return: The start of the period.
    :raise ValueError: When the date is invalid.
    """
    if day is not None:
        return dt.date(int(year), int(month), int(day))
    if month is not None:
        return dt.date(int(year), int(month), 1)
    return dt.date(int(year), 1, 1)


def __get_end(year: str, month: str | None, day: str | None) -> dt.date:
    """Returns the end of the period from the date representation.

    :param year: The year.
    :param month: The month, if any.
    :param day: The day, if any.
    :return: The end of the period.
    :raise ValueError: When the date is invalid.
    """
    if day is not None:
        return dt.date(int(year), int(month), int(day))
    if month is not None:
        year_n: int = int(year)
        month_n: int = int(month)
        day_n: int = calendar.monthrange(year_n, month_n)[1]
        return dt.date(year_n, month_n, day_n)
    return dt.date(int(year), 12, 31)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/4/17

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The unmatched offsets.

"""
import datetime as dt
from decimal import Decimal

from flask import render_template, Response
from flask_babel import LazyString

from accounting.locale import gettext
from accounting.models import Currency, Account, JournalEntryLineItem
from accounting.report.utils.base_page_params import BasePageParams
from accounting.report.utils.base_report import BaseReport
from accounting.report.utils.csv_export import BaseCSVRow, csv_download
from accounting.report.utils.offset_matcher import OffsetMatcher, OffsetPair
from accounting.report.utils.option_link import OptionLink
from accounting.report.utils.report_chooser import ReportChooser
from accounting.report.utils.report_type import ReportType
from accounting.report.utils.unmatched import get_accounts_with_unmatched
from accounting.report.utils.urls import unmatched_url
from accounting.utils.pagination import Pagination


class CSVRow(BaseCSVRow):
    """A row in the CSV."""

    def __init__(self, journal_entry_date: str | dt.date, currency: str,
                 description: str | None, debit: str | Decimal,
                 credit: str | Decimal, balance: str | Decimal):
        """Constructs a row in the CSV.

        :param journal_entry_date: The journal entry date.
        :param currency: The currency.
        :param description: The description.
        :param debit: The debit amount.
        :param credit: The credit amount.
        :param balance: The balance.
        """
        self.date: str | dt.date = journal_entry_date
        """The date."""
        self.currency: str = currency
        """The currency."""
        self.description: str | None = description
        """The description."""
        self.debit: str | Decimal | None = debit
        """The debit amount."""
        self.credit: str | Decimal | None = credit
        """The credit amount."""
        self.balance: str | Decimal = balance
        """The balance."""

    @property
    def values(self) -> list[str | dt.date | Decimal | None]:
        """Returns the values of the row.

        :return: The values of the row.
        """
        return [self.date, self.currency, self.description, self.debit,
                self.credit, self.balance]


class PageParams(BasePageParams):
    """The HTML page parameters."""

    def __init__(self, currency: Currency,
                 account: Account,
                 match_status: str | LazyString,
                 matched_pairs: list[OffsetPair],
                 pagination: Pagination[JournalEntryLineItem],
                 line_items: list[JournalEntryLineItem]):
        """Constructs the HTML page parameters.

        :param currency: The currency.
        :param account: The account.
        :param match_status: The match status message.
        :param matched_pairs: A list of matched pairs.
        :param pagination: The pagination.
        :param line_items: The line items.
        """
        self.currency: Currency = currency
        """The currency."""
        self.account: Account = account
        """The account."""
        self.match_status: str | LazyString = match_status
        """The match status message."""
        self.matched_pairs: list[OffsetPair] = matched_pairs
        """A list of matched pairs."""
        self.pagination: Pagination[JournalEntryLineItem] = pagination
        """The pagination."""
        self.line_items: list[JournalEntryLineItem] = line_items
        """The line items."""

    @property
    def has_data(self) -> bool:
        """Returns whether there is any data on the page.

        :return: True if there is any data, or False otherwise.
        """
        return len(self.line_items) > 0

    @property
    def report_chooser(self) -> ReportChooser:
        """Returns the report chooser.

        :return: The report chooser.
        """
        return ReportChooser(ReportType.UNMATCHED, currency=self.currency,
                             account=self.account)

    @property
    def currency_options(self) -> list[OptionLink]:
        """Returns the currency options.

        :return: The currency options.
        """
        return self._get_currency_options(
            lambda x: unmatched_url(x, self.account), self.currency)

    @property
    def account_options(self) -> list[OptionLink]:
        """Returns the account options.

        :return: The account options.
        """
        options: list[OptionLink] \
            = [OptionLink(gettext("Accounts"),
                          unmatched_url(self.currency, None),
                          False)]
        options.extend(
            [OptionLink(str(x), unmatched_url(self.currency, x),
                        x.id == self.account.id)
             for x in get_accounts_with_unmatched(self.currency)])
        return options


def get_csv_rows(line_items: list[JournalEntryLineItem]) -> list[CSVRow]:
    """Composes and returns the CSV rows from the line items.

    :param line_items: The line items.
    :return: The CSV rows.
    """
    rows: list[CSVRow] = [CSVRow(gettext("Date"), gettext("Currency"),
                                 gettext("Description"), gettext("Debit"),
                                 gettext("Credit"), gettext("Balance"))]
    rows.extend([CSVRow(x.journal_entry.date, x.currency.code,
                        x.description, x.debit, x.credit, x.balance)
                 for x in line_items])
    return rows


class UnmatchedOffsets(BaseReport):
    """The unmatched offsets."""

    def __init__(self, currency: Currency, account: Account):
        """Constructs the unmatched offsets.

        :param currency: The currency.
        :param account: The account.
        """
        self.__currency: Currency = currency
        """The currency."""
        self.__account: Account = account
        """The account."""
        offset_matcher: OffsetMatcher \
            = OffsetMatcher(self.__currency, self.__account)
        self.__line_items: list[JournalEntryLineItem] \
            = offset_matcher.line_items
        """The line items."""
        self.__match_status: str | LazyString = offset_matcher.status
        """The match status message."""
        self.__matched_pairs: list[OffsetPair] = offset_matcher.matched_pairs
        """A list of matched pairs."""

    def csv(self) -> Response:
        """Returns the report as CSV for download.

        :return: The response of the report for download.
        """
        filename: str = "unmatched-{currency}-{account}.csv"\
            .format(currency=self.__currency.code, account=self.__account.code)
        return csv_download(filename, get_csv_rows(self.__line_items))

    def html(self) -> str:
        """Composes and returns the report as HTML.

        :return: The report as HTML.
        """
        pagination: Pagination[JournalEntryLineItem] \
            = Pagination[JournalEntryLineItem](self.__line_items,
                                               is_reversed=True)
        params: PageParams = PageParams(currency=self.__currency,
                                        account=self.__account,
                                        match_status=self.__match_status,
                                        matched_pairs=self.__matched_pairs,
                                        pagination=pagination,
                                        line_items=pagination.list)
        return render_template("accounting/report/unmatched.html",
                               report=params)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/1

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The utility to flash all errors from the forms.

This module should not import any other module from the application.

"""
from typing import Any

from flask import flash
from flask_wtf import FlaskForm


def flash_form_errors(form: FlaskForm) -> None:
    """Flash all errors from a form recursively.

    :param form: The form.
    :return: None.
    """
    __flash_errors(form.errors)


def __flash_errors(error: Any) -> None:
    """Flash all errors recursively.

    :param error: The errors.
    :return: None.
    """
    if isinstance(error, dict):
        for key in error:
            __flash_errors(error[key])
    elif isinstance(error, list):
        for e in error:
            __flash_errors(e)
    else:
        flash(error, "error")



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/4/9

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The test for the reports.

"""
import datetime as dt
import unittest

import httpx
from flask import Flask

from test_site.lib import BaseTestData
from testlib import create_test_app, get_client, get_csrf_token, Accounts

PREFIX: str = "/accounting"
"""The URL prefix for the reports."""
CSV_MIME: str = "text/csv; charset=utf-8"
"""The MIME type of the downloaded CSV files."""


class ReportTestCase(unittest.TestCase):
    """The report test case."""

    def setUp(self) -> None:
        """Sets up the test.
        This is run once per test.

        :return: None.
        """
        self.__app: Flask = create_test_app()
        """The Flask application."""

        with self.__app.app_context():
            from accounting.models import JournalEntry, JournalEntryLineItem
            JournalEntry.query.delete()
            JournalEntryLineItem.query.delete()

        self.__client: httpx.Client = get_client(self.__app, "editor")
        """The user client."""
        self.__csrf_token: str = get_csrf_token(self.__client)
        """The CSRF token."""

    def test_nobody(self) -> None:
        """Test the permission as nobody.

        :return: None.
        """
        client: httpx.Client = get_client(self.__app, "nobody")
        ReportTestData(self.__app, "editor").populate()
        response: httpx.Response

        response = client.get(PREFIX)
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}?as=csv")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/journal")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/journal?as=csv")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/ledger")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/ledger?as=csv")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/income-expenses")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/income-expenses?as=csv")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/trial-balance")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/trial-balance?as=csv")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/income-statement")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/income-statement?as=csv")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/balance-sheet")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/balance-sheet?as=csv")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/unapplied")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/unapplied?as=csv")
        self.assertEqual(response.status_code, 403)

        response = client.get(
            f"{PREFIX}/unapplied/USD/{Accounts.PAYABLE}")
        self.assertEqual(response.status_code, 403)

        response = client.get(
            f"{PREFIX}/unapplied/USD/{Accounts.PAYABLE}?as=csv")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/unmatched")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/unmatched?as=csv")
        self.assertEqual(response.status_code, 403)

        response = client.get(
            f"{PREFIX}/unmatched/USD/{Accounts.PAYABLE}")
        self.assertEqual(response.status_code, 403)

        response = client.get(
            f"{PREFIX}/unmatched/USD/{Accounts.PAYABLE}?as=csv")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/search?q=Salary")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/search?q=Salary&as=csv")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/search?q=薪水")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/search?q=薪水&as=csv")
        self.assertEqual(response.status_code, 403)

    def test_viewer(self) -> None:
        """Test the permission as viewer.

        :return: None.
        """
        client: httpx.Client = get_client(self.__app, "viewer")
        ReportTestData(self.__app, "editor").populate()
        response: httpx.Response

        response = client.get(PREFIX)
        self.assertEqual(response.status_code, 200)

        response = client.get(f"{PREFIX}?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = client.get(f"{PREFIX}/journal")
        self.assertEqual(response.status_code, 200)

        response = client.get(f"{PREFIX}/journal?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = client.get(f"{PREFIX}/ledger")
        self.assertEqual(response.status_code, 200)

        response = client.get(f"{PREFIX}/ledger?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = client.get(f"{PREFIX}/income-expenses")
        self.assertEqual(response.status_code, 200)

        response = client.get(f"{PREFIX}/income-expenses?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = client.get(f"{PREFIX}/trial-balance")
        self.assertEqual(response.status_code, 200)

        response = client.get(f"{PREFIX}/trial-balance?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = client.get(f"{PREFIX}/income-statement")
        self.assertEqual(response.status_code, 200)

        response = client.get(f"{PREFIX}/income-statement?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = client.get(f"{PREFIX}/balance-sheet")
        self.assertEqual(response.status_code, 200)

        response = client.get(f"{PREFIX}/balance-sheet?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = client.get(f"{PREFIX}/unapplied")
        self.assertEqual(response.status_code, 200)

        response = client.get(f"{PREFIX}/unapplied?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = client.get(
            f"{PREFIX}/unapplied/USD/{Accounts.PAYABLE}")
        self.assertEqual(response.status_code, 200)

        response = client.get(
            f"{PREFIX}/unapplied/USD/{Accounts.PAYABLE}?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = client.get(f"{PREFIX}/unmatched")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/unmatched?as=csv")
        self.assertEqual(response.status_code, 403)

        response = client.get(
            f"{PREFIX}/unmatched/USD/{Accounts.PAYABLE}")
        self.assertEqual(response.status_code, 403)

        response = client.get(
            f"{PREFIX}/unmatched/USD/{Accounts.PAYABLE}?as=csv")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/search?q=Salary")
        self.assertEqual(response.status_code, 200)

        response = client.get(f"{PREFIX}/search?q=Salary&as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = client.get(f"{PREFIX}/search?q=薪水")
        self.assertEqual(response.status_code, 200)

        response = client.get(f"{PREFIX}/search?q=薪水&as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

    def test_editor(self) -> None:
        """Test the permission as editor.

        :return: None.
        """
        ReportTestData(self.__app, "editor").populate()
        response: httpx.Response

        response = self.__client.get(PREFIX)
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/journal")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/journal?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/ledger")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/ledger?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/income-expenses")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/income-expenses?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/trial-balance")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/trial-balance?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/income-statement")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/income-statement?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/balance-sheet")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/balance-sheet?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/unapplied")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/unapplied?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(
            f"{PREFIX}/unapplied/USD/{Accounts.PAYABLE}")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(
            f"{PREFIX}/unapplied/USD/{Accounts.PAYABLE}?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/unmatched")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/unmatched?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(
            f"{PREFIX}/unmatched/USD/{Accounts.PAYABLE}")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(
            f"{PREFIX}/unmatched/USD/{Accounts.PAYABLE}?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/search?q=Salary")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/search?q=Salary&as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/search?q=薪水")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/search?q=薪水&as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

    def test_empty_db(self) -> None:
        """Tests the empty database.

        :return: None.
        """
        response: httpx.Response

        response = self.__client.get(PREFIX)
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/journal")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/journal?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/ledger")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/ledger?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/income-expenses")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/income-expenses?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/trial-balance")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/trial-balance?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/income-statement")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/income-statement?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/balance-sheet")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/balance-sheet?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/unapplied")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/unapplied?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(
            f"{PREFIX}/unapplied/USD/{Accounts.PAYABLE}")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(
            f"{PREFIX}/unapplied/USD/{Accounts.PAYABLE}?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/unmatched")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/unmatched?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(
            f"{PREFIX}/unmatched/USD/{Accounts.PAYABLE}")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(
            f"{PREFIX}/unmatched/USD/{Accounts.PAYABLE}?as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)

        response = self.__client.get(f"{PREFIX}/search?q=Salary")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/search?q=Salary&as=csv")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], CSV_MIME)


class ReportTestData(BaseTestData):
    """The report test data."""

    def _init_data(self) -> None:
        today: dt.date = dt.date.today()
        year: int = today.year - 5
        month: int = today.month
        while True:
            date: dt.date = dt.date(year, month, 5)
            if date > today:
                break
            self._add_simple_journal_entry(
                (date - today).days, "USD",
                "Salary薪水", "1200", Accounts.BANK, Accounts.SERVICE)
            month = month + 1
            if month > 12:
                year = year + 1
                month = 1
        self._add_simple_journal_entry(
            1, "USD", "Withdraw領錢", "1000", Accounts.CASH, Accounts.BANK)
        self._add_simple_journal_entry(
            0, "USD", "Dinner晚餐", "40", Accounts.MEAL, Accounts.CASH)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/7

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The trial balance.

"""
from decimal import Decimal

import sqlalchemy as sa
from flask import Response, render_template

from accounting import db
from accounting.locale import gettext
from accounting.models import Currency, Account, JournalEntry, \
    JournalEntryLineItem
from accounting.report.period import Period, PeriodChooser
from accounting.report.utils.base_page_params import BasePageParams
from accounting.report.utils.base_report import BaseReport
from accounting.report.utils.csv_export import BaseCSVRow, csv_download, \
    period_spec
from accounting.report.utils.option_link import OptionLink
from accounting.report.utils.report_chooser import ReportChooser
from accounting.report.utils.report_type import ReportType
from accounting.report.utils.urls import ledger_url, trial_balance_url


class ReportAccount:
    """An account in the report."""

    def __init__(self, account: Account, amount: Decimal, url: str):
        """Constructs an account in the report.

        :param account: The account.
        :param amount: The amount.
        :param url: The URL to the ledger of the account.
        """
        self.account: Account = account
        """The account."""
        self.debit: Decimal | None = amount if amount > 0 else None
        """The debit amount."""
        self.credit: Decimal | None = -amount if amount < 0 else None
        """The credit amount."""
        self.url: str = url
        """The URL to the ledger of the account."""


class Total:
    """The totals."""

    def __init__(self, debit: Decimal, credit: Decimal):
        """Constructs the total in the trial balance.

        :param debit: The debit amount.
        :param credit: The credit amount.
        """
        self.debit: Decimal | None = debit
        """The debit amount."""
        self.credit: Decimal | None = credit
        """The credit amount."""


class CSVRow(BaseCSVRow):
    """A row in the CSV."""

    def __init__(self, text: str | None,
                 debit: str | Decimal | None,
                 credit: str | Decimal | None):
        """Constructs a row in the CSV.

        :param text: The text.
        :param debit: The debit amount.
        :param credit: The credit amount.
        """
        self.text: str | None = text
        """The text."""
        self.debit: str | Decimal | None = debit
        """The debit amount."""
        self.credit: str | Decimal | None = credit
        """The credit amount."""

    @property
    def values(self) -> list[str | Decimal | None]:
        """Returns the values of the row.

        :return: The values of the row.
        """
        return [self.text, self.debit, self.credit]


class PageParams(BasePageParams):
    """The HTML page parameters."""

    def __init__(self, currency: Currency,
                 period: Period,
                 accounts: list[ReportAccount],
                 total: Total):
        """Constructs the HTML page parameters.

        :param currency: The currency.
        :param period: The period.
        :param accounts: The accounts in the trial balance.
        :param total: The total of the trial balance.
        """
        self.currency: Currency = currency
        """The currency."""
        self.period: Period = period
        """The period."""
        self.accounts: list[ReportAccount] = accounts
        """The accounts in the trial balance."""
        self.total: Total = total
        """The total of the trial balance."""
        self.period_chooser: PeriodChooser = PeriodChooser(
            lambda x: trial_balance_url(currency, x))
        """The period chooser."""

    @property
    def has_data(self) -> bool:
        """Returns whether there is any data on the page.

        :return: True if there is any data, or False otherwise.
        """
        return len(self.accounts) > 0

    @property
    def report_chooser(self) -> ReportChooser:
        """Returns the report chooser.

        :return: The report chooser.
        """
        return ReportChooser(ReportType.TRIAL_BALANCE,
                             currency=self.currency,
                             period=self.period)

    @property
    def currency_options(self) -> list[OptionLink]:
        """Returns the currency options.

        :return: The currency options.
        """
        return self._get_currency_options(
            lambda x: trial_balance_url(x, self.period), self.currency)


class TrialBalance(BaseReport):
    """The trial balance."""

    def __init__(self, currency: Currency, period: Period):
        """Constructs a trial balance.

        :param currency: The currency.
        :param period: The period.
        """
        self.__currency: Currency = currency
        """The currency."""
        self.__period: Period = period
        """The period."""
        self.__accounts: list[ReportAccount]
        """The accounts in the trial balance."""
        self.__total: Total
        """The total of the trial balance."""
        self.__set_data()

    def __set_data(self) -> None:
        """Queries and sets data sections in the trial balance.

        :return: None.
        """
        conditions: list[sa.BinaryExpression] \
            = [JournalEntryLineItem.currency_code == self.__currency.code]
        if self.__period.start is not None:
            conditions.append(JournalEntry.date >= self.__period.start)
        if self.__period.end is not None:
            conditions.append(JournalEntry.date <= self.__period.end)
        balance_func: sa.Function = sa.func.sum(sa.case(
            (JournalEntryLineItem.is_debit, JournalEntryLineItem.amount),
            else_=-JournalEntryLineItem.amount)).label("balance")
        select_balances: sa.Select = sa.select(Account.id, balance_func)\
            .join(JournalEntry).join(Account)\
            .filter(*conditions)\
            .group_by(Account.id)\
            .having(balance_func != 0)\
            .order_by(Account.base_code, Account.no)
        balances: list[sa.Row] = db.session.execute(select_balances).all()
        accounts: dict[int, Account] \
            = {x.id: x for x in Account.query
               .filter(Account.id.in_([x.id for x in balances])).all()}
        self.__accounts = [ReportAccount(account=accounts[x.id],
                                         amount=x.balance,
                                         url=ledger_url(self.__currency,
                                                        accounts[x.id],
                                                        self.__period))
                           for x in balances]
        self.__total = Total(
            sum([x.debit for x in self.__accounts if x.debit is not None]),
            sum([x.credit for x in self.__accounts if x.credit is not None]))

    def csv(self) -> Response:
        """Returns the report as CSV for download.

        :return: The response of the report for download.
        """
        filename: str = "trial-balance-{currency}-{period}.csv"\
            .format(currency=self.__currency.code,
                    period=period_spec(self.__period))
        return csv_download(filename, self.__get_csv_rows())

    def __get_csv_rows(self) -> list[CSVRow]:
        """Composes and returns the CSV rows.

        :return: The CSV rows.
        """
        rows: list[CSVRow] = [CSVRow(gettext("Account"), gettext("Debit"),
                                     gettext("Credit"))]
        rows.extend([CSVRow(str(x.account), x.debit, x.credit)
                     for x in self.__accounts])
        rows.append(CSVRow(gettext("Total"), self.__total.debit,
                           self.__total.credit))
        return rows

    def html(self) -> str:
        """Composes and returns the report as HTML.

        :return: The report as HTML.
        """
        params: PageParams = PageParams(currency=self.__currency,
                                        period=self.__period,
                                        accounts=self.__accounts,
                                        total=self.__total)
        return render_template("accounting/report/trial-balance.html",
                               report=params)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/6

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The page parameters of a report.

"""
from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Type
from urllib.parse import urlparse, ParseResult, parse_qsl, urlencode, \
    urlunparse

import sqlalchemy as sa
from flask import request

from accounting import db
from accounting.models import Currency, JournalEntryLineItem
from accounting.utils.journal_entry_types import JournalEntryType
from .option_link import OptionLink
from .report_chooser import ReportChooser


class BasePageParams(ABC):
    """The base HTML page parameters class."""

    @property
    @abstractmethod
    def has_data(self) -> bool:
        """Returns whether there is any data on the page.

        :return: True if there is any data, or False otherwise.
        """

    @property
    @abstractmethod
    def report_chooser(self) -> ReportChooser:
        """Returns the report chooser.

        :return: The report chooser.
        """

    @property
    def journal_entry_types(self) -> Type[JournalEntryType]:
        """Returns the journal entry types.

        :return: The journal entry types.
        """
        return JournalEntryType

    @property
    def csv_uri(self) -> str:
        uri: str = request.full_path if request.query_string \
            else request.path
        uri_p: ParseResult = urlparse(uri)
        params: list[tuple[str, str]] = parse_qsl(uri_p.query)
        params = [x for x in params if x[0] != "as"]
        params.append(("as", "csv"))
        parts: list[str] = list(uri_p)
        parts[4] = urlencode(params)
        return urlunparse(parts)

    @staticmethod
    def _get_currency_options(get_url: Callable[[Currency], str],
                              active_currency: Currency) -> list[OptionLink]:
        """Returns the currency options.

        :param get_url: The callback to return the URL of a currency.
        :param active_currency: The active currency.
        :return: The currency options.
        """
        in_use: set[str] = set(db.session.scalars(
            sa.select(JournalEntryLineItem.currency_code)
            .group_by(JournalEntryLineItem.currency_code)).all())
        return [OptionLink(str(x), get_url(x), x.code == active_currency.code)
                for x in Currency.query.filter(Currency.code.in_(in_use))
                .order_by(Currency.code).all()]



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/22

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The test for the options.

"""
import datetime as dt
import unittest

import httpx
from flask import Flask

from accounting.utils.next_uri import encode_next
from test_site import db
from testlib import NEXT_URI, Accounts, create_test_app, get_client, \
    get_csrf_token

PREFIX: str = "/accounting/options"
"""The URL prefix for the option management."""


class OptionTestCase(unittest.TestCase):
    """The option test case."""

    def setUp(self) -> None:
        """Sets up the test.
        This is run once per test.

        :return: None.
        """
        self.__app: Flask = create_test_app()
        """The Flask application."""

        with self.__app.app_context():
            from accounting.models import Option
            Option.query.delete()
            self.__encoded_next_uri: str = encode_next(NEXT_URI)
            """The encoded next URI."""

        self.__client: httpx.Client = get_client(self.__app, "admin")
        """The user client."""
        self.__csrf_token: str = get_csrf_token(self.__client)
        """The CSRF token."""

    def test_nobody(self) -> None:
        """Test the permission as nobody.

        :return: None.
        """
        client: httpx.Client = get_client(self.__app, "nobody")
        csrf_token: str = get_csrf_token(client)
        detail_uri: str = f"{PREFIX}?next={self.__encoded_next_uri}"
        edit_uri: str = f"{PREFIX}/edit?next={self.__encoded_next_uri}"
        update_uri: str = f"{PREFIX}/update"
        response: httpx.Response

        response = client.get(detail_uri)
        self.assertEqual(response.status_code, 403)

        response = client.get(edit_uri)
        self.assertEqual(response.status_code, 403)

        response = client.post(update_uri, data=self.__get_form(csrf_token))
        self.assertEqual(response.status_code, 403)

    def test_viewer(self) -> None:
        """Test the permission as viewer.

        :return: None.
        """
        client: httpx.Client = get_client(self.__app, "viewer")
        csrf_token: str = get_csrf_token(client)
        detail_uri: str = f"{PREFIX}?next={self.__encoded_next_uri}"
        edit_uri: str = f"{PREFIX}/edit?next={self.__encoded_next_uri}"
        update_uri: str = f"{PREFIX}/update"
        response: httpx.Response

        response = client.get(detail_uri)
        self.assertEqual(response.status_code, 403)

        response = client.get(edit_uri)
        self.assertEqual(response.status_code, 403)

        response = client.post(update_uri, data=self.__get_form(csrf_token))
        self.assertEqual(response.status_code, 403)

    def test_editor(self) -> None:
        """Test the permission as editor.

        :return: None.
        """
        client: httpx.Client = get_client(self.__app, "editor")
        csrf_token: str = get_csrf_token(client)
        detail_uri: str = f"{PREFIX}?next={self.__encoded_next_uri}"
        edit_uri: str = f"{PREFIX}/edit?next={self.__encoded_next_uri}"
        update_uri: str = f"{PREFIX}/update"
        response: httpx.Response

        response = client.get(detail_uri)
        self.assertEqual(response.status_code, 403)

        response = client.get(edit_uri)
        self.assertEqual(response.status_code, 403)

        response = client.post(update_uri, data=self.__get_form(csrf_token))
        self.assertEqual(response.status_code, 403)

    def test_admin(self) -> None:
        """Test the permission as administrator.

        :return: None.
        """
        detail_uri: str = f"{PREFIX}?next={self.__encoded_next_uri}"
        edit_uri: str = f"{PREFIX}/edit?next={self.__encoded_next_uri}"
        update_uri: str = f"{PREFIX}/update"
        response: httpx.Response

        response = self.__client.get(detail_uri)
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(edit_uri)
        self.assertEqual(response.status_code, 200)

        response = self.__client.post(update_uri, data=self.__get_form())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

    def test_set(self) -> None:
        """Test to set the options.

        :return: None.
        """
        from accounting.utils.options import options
        detail_uri: str = f"{PREFIX}?next={self.__encoded_next_uri}"
        edit_uri: str = f"{PREFIX}/edit?next={self.__encoded_next_uri}"
        update_uri: str = f"{PREFIX}/update"
        form: dict[str, str]
        response: httpx.Response

        # Empty currency code
        form = self.__get_form()
        form["default_currency_code"] = " "
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Non-existing currency code
        form = self.__get_form()
        form["default_currency_code"] = "ZZZ"
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Empty current account
        form = self.__get_form()
        form["default_ie_account_code"] = " "
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Non-existing current account
        form = self.__get_form()
        form["default_ie_account_code"] = "9999-999"
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not a current account
        form = self.__get_form()
        form["default_ie_account_code"] = Accounts.MEAL
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Recurring item name empty
        form = self.__get_form()
        key = [x for x in form if x.endswith("-name")][0]
        form[key] = " "
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Recurring item account empty
        form = self.__get_form()
        key = [x for x in form if x.endswith("-account_code")][0]
        form[key] = " "
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Recurring item non-expense account
        form = self.__get_form()
        key = [x for x in form
               if x.startswith("recurring-expense-")
               and x.endswith("-account_code")][0]
        form[key] = Accounts.SERVICE
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Recurring item non-income account
        form = self.__get_form()
        key = [x for x in form
               if x.startswith("recurring-income-")
               and x.endswith("-account_code")][0]
        form[key] = Accounts.UTILITIES
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Recurring item payable expense
        form = self.__get_form()
        key = [x for x in form
               if x.startswith("recurring-expense-")
               and x.endswith("-account_code")][0]
        form[key] = Accounts.PAYABLE
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Recurring item receivable income
        form = self.__get_form()
        key = [x for x in form
               if x.startswith("recurring-income-")
               and x.endswith("-account_code")][0]
        form[key] = Accounts.RECEIVABLE
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Recurring item description template empty
        form = self.__get_form()
        key = [x for x in form if x.endswith("-description_template")][0]
        form[key] = " "
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Success, with malformed order
        with self.__app.app_context():
            self.assertEqual(options.default_currency_code, "USD")
            self.assertEqual(options.default_ie_account_code, "1111-001")
            self.assertEqual(len(options.recurring.expenses), 0)
            self.assertEqual(len(options.recurring.incomes), 0)

        response = self.__client.post(update_uri, data=self.__get_form())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.__app.app_context():
            self.assertEqual(options.default_currency_code, "EUR")
            self.assertEqual(options.default_ie_account_code, "0000-000")
            self.assertEqual(len(options.recurring.expenses), 4)
            self.assertEqual(options.recurring.expenses[0].account_code,
                             Accounts.INSURANCE)
            self.assertEqual(options.recurring.expenses[1].account_code,
                             Accounts.POSTAGE)
            self.assertEqual(options.recurring.expenses[2].account_code,
                             Accounts.UTILITIES)
            self.assertEqual(options.recurring.expenses[3].account_code,
                             Accounts.RENT_EXPENSE)
            self.assertEqual(len(options.recurring.incomes), 2)
            self.assertEqual(options.recurring.incomes[0].account_code,
                             Accounts.SERVICE)
            self.assertEqual(options.recurring.incomes[1].account_code,
                             Accounts.DONATION)

        # Success, with no recurring data
        form = self.__get_form()
        form = {x: form[x] for x in form if not x.startswith("recurring-")}
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.__app.app_context():
            self.assertEqual(len(options.recurring.expenses), 0)
            self.assertEqual(len(options.recurring.incomes), 0)

    def test_update_not_modified(self) -> None:
        """Tests that the data is not modified.

        :return: None.
        """
        from accounting.models import Option
        detail_uri: str = f"{PREFIX}?next={self.__encoded_next_uri}"
        update_uri: str = f"{PREFIX}/update"
        form: dict[str, str]
        option: Option | None
        resource: httpx.Response

        response = self.__client.post(update_uri, data=self.__get_form())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.__app.app_context():
            option = db.session.get(Option, "recurring")
            self.assertIsNotNone(option)
            timestamp: dt.datetime \
                = option.created_at - dt.timedelta(seconds=5)
            option.created_at = timestamp
            option.updated_at = timestamp
            db.session.commit()

        # The recurring setting was not modified
        form = self.__get_form()
        form["default_currency_code"] = "JPY"
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.__app.app_context():
            option = db.session.get(Option, "recurring")
            self.assertIsNotNone(option)
            self.assertEqual(option.created_at, timestamp)
            self.assertEqual(option.updated_at, timestamp)

        # The recurring setting was modified
        form = self.__get_form()
        key: str = [x for x in form
                    if x.startswith("recurring-expense-")
                    and x.endswith("-account_code")][0]
        form[key] = Accounts.MEAL
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.__app.app_context():
            option = db.session.get(Option, "recurring")
            self.assertIsNotNone(option)
            self.assertLess(option.created_at, option.updated_at)

    def test_created_updated_by(self) -> None:
        """Tests the created-by and updated-by record.

        :return: None.
        """
        from accounting.models import Option
        from accounting.utils.user import get_user_pk
        admin_username, editor_username = "admin", "editor"
        detail_uri: str = f"{PREFIX}?next={self.__encoded_next_uri}"
        update_uri: str = f"{PREFIX}/update"
        option: Option | None
        response: httpx.Response

        response = self.__client.post(update_uri, data=self.__get_form())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.__app.app_context():
            editor_pk: int = get_user_pk(editor_username)
            option = db.session.get(Option, "recurring")
            self.assertIsNotNone(option)
            option.created_by_id = editor_pk
            option.updated_by_id = editor_pk
            db.session.commit()

        form: dict[str, str] = self.__get_form()
        key: str = [x for x in form
                    if x.startswith("recurring-expense-")
                    and x.endswith("-account_code")][0]
        form[key] = Accounts.MEAL
        response = self.__client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.__app.app_context():
            option = db.session.get(Option, "recurring")
            self.assertIsNotNone(option)
            self.assertEqual(option.created_by.username, editor_username)
            self.assertEqual(option.updated_by.username, admin_username)

    def __get_form(self, csrf_token: str | None = None) -> dict[str, str]:
        """Returns the option form.

        :param csrf_token: The CSRF token.
        :return: The option form.
        """
        if csrf_token is None:
            csrf_token = self.__csrf_token
        return {"csrf_token": csrf_token,
                "next": self.__encoded_next_uri,
                "default_currency_code": "EUR",
                "default_ie_account_code": "0000-000",
                "recurring-expense-1-name": "Water bill",
                "recurring-expense-1-account_code": Accounts.UTILITIES,
                "recurring-expense-1-description_template":
                    "Water bill for {last_bimonthly_name}",
                "recurring-expense-3-no": "16",
                "recurring-expense-3-name": "Phone bill",
                "recurring-expense-3-account_code": Accounts.POSTAGE,
                "recurring-expense-3-description_template":
                    "Phone bill for {last_month_name}",
                "recurring-expense-12-name": "Rent",
                "recurring-expense-12-account_code": Accounts.RENT_EXPENSE,
                "recurring-expense-12-description_template":
                    "Rent for {this_month_name}",
                "recurring-expense-26-no": "7",
                "recurring-expense-26-name": "Insurance",
                "recurring-expense-26-account_code": Accounts.INSURANCE,
                "recurring-expense-26-description_template":
                    "Insurance for {last_month_name}",
                "recurring-income-2-no": "12",
                "recurring-income-2-name": "Donation",
                "recurring-income-2-account_code": Accounts.DONATION,
                "recurring-income-2-description_template":
                    "Donation for {this_month_name}",
                "recurring-income-15-no": "4",
                "recurring-income-15-name": "Payroll",
                "recurring-income-15-account_code": Accounts.SERVICE,
                "recurring-income-15-description_template":
                    "Payroll for {last_month_name}"}



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/1

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The text stripper for the form fields.

This module should not import any other module from the application.

"""
import re


def strip_text(s: str | None) -> str | None:
    """The filter to strip the leading and trailing white spaces of text.

    :param s: The text input string.
    :return: The filtered string.
    """
    if s is None:
        return None
    s = s.strip()
    return s if s != "" else None


def strip_multiline_text(s: str | None) -> str | None:
    """The filter to strip a piece of multi-line text.

    :param s: The text input string.
    :return: The filtered string.
    """
    if s is None:
        return None
    s = re.sub(r"^\s*\n", "", s.rstrip())
    return s if s != "" else None



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/4/10

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The test for the console commands.

"""
import csv
import datetime as dt
import re
import unittest
from typing import Any

import sqlalchemy as sa
from click.testing import Result
from flask import Flask
from flask.testing import FlaskCliRunner
from sqlalchemy.sql.ddl import DropTable

from test_site import db
from testlib import create_test_app


class ConsoleCommandTestCase(unittest.TestCase):
    """The console command test case."""

    def setUp(self) -> None:
        """Sets up the test.
        This is run once per test.

        :return: None.
        """
        self.__app: Flask = create_test_app()
        """The Flask application."""

    def test_init_db(self) -> None:
        """Tests the "accounting-init-db" console command.

        :return: None.
        """
        with self.__app.app_context():
            # Drop every accounting table, to see if accounting-init-db
            # recreates them correctly.
            tables: list[sa.Table] \
                = [db.metadata.tables[x] for x in db.metadata.tables
                   if x.startswith("accounting_")]
            for table in tables:
                db.session.execute(DropTable(table))
            db.session.commit()
            inspector: sa.Inspector = sa.inspect(db.session.connection())
            self.assertEqual(len({x for x in inspector.get_table_names()
                                  if x.startswith("accounting_")}),
                             0)

        runner: FlaskCliRunner = self.__app.test_cli_runner()
        with self.__app.app_context():
            result: Result = runner.invoke(
                args=["accounting-init-db", "-u", "editor"])
        self.assertEqual(result.exit_code, 0,
                         result.output + str(result.exception))
        self.__test_base_account_data()
        self.__test_account_data()
        self.__test_currency_data()

    def __test_base_account_data(self) -> None:
        """Tests the base account data.

        :return: None.
        """
        from accounting import data_dir
        from accounting.models import BaseAccount

        with open(data_dir / "base_accounts.csv") as fp:
            rows: list[dict[str, str]] = list(csv.DictReader(fp))
        data: dict[dict[str, Any]] \
            = {x["code"]: {"code": x["code"],
                           "title": x["title"],
                           "l10n": {y[5:]: x[y]
                                    for y in x if y.startswith("l10n-")}}
               for x in rows}

        with self.__app.app_context():
            accounts: list[BaseAccount] = BaseAccount.query.all()

        self.assertEqual(len(accounts), len(data))
        for account in accounts:
            self.assertIn(account.code, data)
            self.assertEqual(account.title_l10n.lower(),
                             data[account.code]["title"].lower())
            self.__test_title_case(account.title_l10n)
            l10n: dict[str, str] = {x.locale: x.title for x in account.l10n}
            self.assertEqual(len(l10n), len(data[account.code]["l10n"]))
            for locale in l10n:
                self.assertIn(locale, data[account.code]["l10n"])
                self.assertEqual(l10n[locale],
                                 data[account.code]["l10n"][locale])

    def __test_title_case(self, s: str) -> None:
        """Tests the case of a base account title.

        :param s: The base account title.
        :return: None.
        """
        from accounting.utils.title_case import MINOR_WORDS

        self.assertTrue(s[0].isupper(), s)
        for word in re.findall(r"\w+", s):
            if len(word) >= 4:
                self.assertTrue(word.istitle(), s)
            elif word in MINOR_WORDS:
                self.assertTrue(word.islower(), s)
            else:
                self.assertTrue(word.istitle(), s)

    def __test_account_data(self) -> None:
        """Tests the account data.

        :return: None.
        """
        from accounting.models import BaseAccount, Account, AccountL10n

        with self.__app.app_context():
            bases: list[BaseAccount] = BaseAccount.query\
                .filter(sa.func.char_length(BaseAccount.code) == 4).all()
            accounts: list[Account] = Account.query.all()
            l10n: list[AccountL10n] = AccountL10n.query.all()

        self.assertEqual({x.code for x in bases},
                         {x.base_code for x in accounts})
        self.assertEqual(len(accounts), len(bases))
        self.assertEqual(len(l10n), len(bases) * 2)
        base_dict: dict[str, BaseAccount] = {x.code: x for x in bases}
        for account in accounts:
            base: BaseAccount = base_dict[account.base_code]
            self.assertEqual(account.no, 1)
            self.assertEqual(account.title_l10n, base.title_l10n)
            self.assertEqual({x.locale: x.title for x in account.l10n},
                             {x.locale: x.title for x in base.l10n})

    def __test_currency_data(self) -> None:
        """Tests the currency data.

        :return: None.
        """
        from accounting import data_dir
        from accounting.models import Currency

        with open(data_dir / "currencies.csv") as fp:
            data: dict[dict[str, Any]] \
                = {x["code"]: {"code": x["code"],
                               "name": x["name"],
                               "l10n": {y[5:]: x[y]
                                        for y in x if y.startswith("l10n-")}}
                   for x in csv.DictReader(fp)}

        with self.__app.app_context():
            currencies: list[Currency] = Currency.query.all()

        self.assertEqual(len(currencies), len(data))
        for currency in currencies:
            self.assertIn(currency.code, data)
            self.assertEqual(currency.name_l10n, data[currency.code]["name"])
            l10n: dict[str, str] = {x.locale: x.name for x in currency.l10n}
            self.assertEqual(len(l10n), len(data[currency.code]["l10n"]))
            for locale in l10n:
                self.assertIn(locale, data[currency.code]["l10n"])
                self.assertEqual(l10n[locale],
                                 data[currency.code]["l10n"][locale])

    def test_titleize(self) -> None:
        """Tests the "accounting-titleize" console command.

        :return: None.
        """
        from accounting.models import BaseAccount, Account
        from accounting.utils.random_id import new_id
        from accounting.utils.user import get_user_pk
        runner: FlaskCliRunner = self.__app.test_cli_runner()

        with self.__app.app_context():
            # Resets the accounts.
            tables: list[sa.Table] \
                = [db.metadata.tables[x] for x in db.metadata.tables
                   if x.startswith("accounting_")]
            for table in tables:
                db.session.execute(DropTable(table))
            db.session.commit()
            inspector: sa.Inspector = sa.inspect(db.session.connection())
            self.assertEqual(len({x for x in inspector.get_table_names()
                                  if x.startswith("accounting_")}),
                             0)
            result: Result = runner.invoke(
                args=["accounting-init-db", "-u", "editor"])
            self.assertEqual(result.exit_code, 0,
                             result.output + str(result.exception))

            # Turns the titles into lowercase.
            for base in BaseAccount.query:
                base.title_l10n = base.title_l10n.lower()
            for account in Account.query:
                account.title_l10n = account.title_l10n.lower()
                account.created_at \
                    = account.created_at - dt.timedelta(seconds=5)
                account.updated_at = account.created_at

            # Adds a custom account.
            custom_title = "MBK Bank"
            creator_pk: int = get_user_pk("editor")
            new_account: Account = Account(
                id=new_id(Account),
                base_code="1112",
                no="2",
                title_l10n=custom_title,
                is_need_offset=False,
                created_by_id=creator_pk,
                updated_by_id=creator_pk)
            db.session.add(new_account)
            db.session.commit()

            result: Result = runner.invoke(
                args=["accounting-titleize", "-u", "editor"])
            self.assertEqual(result.exit_code, 0,
                             result.output + str(result.exception))
            for base in BaseAccount.query:
                self.__test_title_case(base.title_l10n)
            for account in Account.query:
                if account.id != new_account.id:
                    self.__test_title_case(account.title_l10n)
                    self.assertNotEqual(account.created_at, account.updated_at)
                else:
                    self.assertEqual(account.title_l10n, custom_title)

            db.session.delete(new_account)
            db.session.commit()



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/4/7

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The accounts with unapplied original line items.

"""
import datetime as dt
from decimal import Decimal

from flask import render_template, Response

from accounting.locale import gettext
from accounting.models import Currency, Account
from accounting.report.utils.base_page_params import BasePageParams
from accounting.report.utils.base_report import BaseReport
from accounting.report.utils.csv_export import BaseCSVRow, csv_download
from accounting.report.utils.option_link import OptionLink
from accounting.report.utils.report_chooser import ReportChooser
from accounting.report.utils.report_type import ReportType
from accounting.report.utils.unapplied import get_accounts_with_unapplied
from accounting.report.utils.urls import unapplied_url


class CSVRow(BaseCSVRow):
    """A row in the CSV."""

    def __init__(self, account: str, count: int | str):
        """Constructs a row in the CSV.

        :param account: The account.
        :param count: The number of unapplied original line items.
        """
        self.account: str = account
        """The currency."""
        self.count: int | str = count
        """The number of unapplied original line items."""

    @property
    def values(self) -> list[str | dt.date | Decimal | None]:
        """Returns the values of the row.

        :return: The values of the row.
        """
        return [self.account, self.count]


class PageParams(BasePageParams):
    """The HTML page parameters."""

    def __init__(self, currency: Currency, accounts: list[Account]):
        """Constructs the HTML page parameters.

        :param currency: The currency.
        :param accounts: The accounts.
        """
        self.currency: Currency = currency
        """The currency."""
        self.accounts: list[Account] = accounts
        """The accounts."""

    @property
    def has_data(self) -> bool:
        """Returns whether there is any data on the page.

        :return: True if there is any data, or False otherwise.
        """
        return len(self.accounts) > 0

    @property
    def report_chooser(self) -> ReportChooser:
        """Returns the report chooser.

        :return: The report chooser.
        """
        return ReportChooser(ReportType.UNAPPLIED, currency=self.currency,
                             account=None)

    @property
    def currency_options(self) -> list[OptionLink]:
        """Returns the currency options.

        :return: The currency options.
        """
        return self._get_currency_options(lambda x: unapplied_url(x, None),
                                          self.currency)

    @property
    def account_options(self) -> list[OptionLink]:
        """Returns the account options.

        :return: The account options.
        """
        options: list[OptionLink] \
            = [OptionLink(gettext("Accounts"),
                          unapplied_url(self.currency, None),
                          True)]
        options.extend(
            [OptionLink(str(x), unapplied_url(self.currency, x), False)
             for x in self.accounts])
        return options


def get_csv_rows(accounts: list[Account]) -> list[CSVRow]:
    """Composes and returns the CSV rows from the line items.

    :param accounts: The accounts.
    :return: The CSV rows.
    """
    rows: list[CSVRow] = [CSVRow(gettext("Account"), gettext("Count"))]
    rows.extend([CSVRow(str(x), x.count) for x in accounts])
    return rows


class AccountsWithUnappliedOriginalLineItems(BaseReport):
    """The accounts with unapplied original line items."""

    def __init__(self, currency: Currency):
        """Constructs the outstanding balances.

        :param currency: The currency.
        """
        self.__currency: Currency = currency
        """The currency."""
        self.__accounts: list[Account] = get_accounts_with_unapplied(currency)
        """The accounts."""

    def csv(self) -> Response:
        """Returns the report as CSV for download.

        :return: The response of the report for download.
        """
        filename: str = "unapplied-accounts.csv"
        return csv_download(filename, get_csv_rows(self.__accounts))

    def html(self) -> str:
        """Composes and returns the report as HTML.

        :return: The report as HTML.
        """
        return render_template("accounting/report/unapplied-accounts.html",
                               report=PageParams(currency=self.__currency,
                                                 accounts=self.__accounts))



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/1

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The forms for the account management.

"""
import sqlalchemy as sa
from flask import request
from flask_wtf import FlaskForm
from wtforms import StringField, BooleanField
from wtforms.validators import DataRequired, ValidationError

from accounting import db
from accounting.locale import lazy_gettext
from accounting.models import BaseAccount, Account
from accounting.utils.random_id import new_id
from accounting.utils.strip_text import strip_text
from accounting.utils.user import get_current_user_pk


class BaseAccountExists:
    """The validator to check if the base account exists."""

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        if field.data == "":
            return
        if db.session.get(BaseAccount, field.data) is None:
            raise ValidationError(lazy_gettext(
                "The base account does not exist."))


class BaseAccountAvailable:
    """The validator to check if the base account is available."""

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        if field.data == "":
            return
        if len(field.data) != 4:
            raise ValidationError(lazy_gettext(
                "The base account is not available."))


class NoOffsetNominalAccount:
    """The validator to check nominal account is not to be offset."""

    def __call__(self, form: FlaskForm, field: BooleanField) -> None:
        assert isinstance(form, AccountForm)
        if not field.data:
            return
        if form.base_code.data is None:
            return
        if form.base_code.data[0] not in {"1", "2", "3"}:
            raise ValidationError(lazy_gettext(
                "A nominal account does not need offset."))


class AccountForm(FlaskForm):
    """The form to create or edit an account."""
    base_code = StringField(
        filters=[strip_text],
        validators=[
            DataRequired(lazy_gettext("Please select the base account.")),
            BaseAccountExists(),
            BaseAccountAvailable()])
    """The code of the base account."""
    title = StringField(
        filters=[strip_text],
        validators=[DataRequired(lazy_gettext("Please fill in the title"))])
    """The title."""
    is_need_offset = BooleanField(
        validators=[NoOffsetNominalAccount()])
    """Whether the the journal entry line items of this account need offset."""

    def populate_obj(self, obj: Account) -> None:
        """Populates the form data into an account object.

        :param obj: The account object.
        :return: None.
        """
        is_new: bool = obj.id is None
        if is_new:
            obj.id = new_id(Account)
        if obj.base_code != self.base_code.data:
            if obj.base_code is not None:
                sort_accounts_in(obj.base_code, obj.id)
            sort_accounts_in(self.base_code.data, obj.id)
            count: int = Account.query\
                .filter(Account.base_code == self.base_code.data).count()
            obj.base_code = self.base_code.data
            obj.no = count + 1
        obj.title = self.title.data
        if self.base_code.data[0] in {"1", "2", "3"}:
            obj.is_need_offset = self.is_need_offset.data
        else:
            obj.is_need_offset = False
        if is_new:
            current_user_pk: int = get_current_user_pk()
            obj.created_by_id = current_user_pk
            obj.updated_by_id = current_user_pk

    def post_update(self, obj: Account) -> None:
        """The post-processing after the update.

        :param obj: The account object.
        :return: None
        """
        current_user_pk: int = get_current_user_pk()
        obj.updated_by_id = current_user_pk
        obj.updated_at = sa.func.now()
        if hasattr(self, "__post_update"):
            getattr(self, "__post_update")()

    @property
    def selected_base(self) -> BaseAccount | None:
        """The selected base account in the form.

        :return: The selected base account in the form.
        """
        return db.session.get(BaseAccount, self.base_code.data)

    @property
    def base_options(self) -> list[BaseAccount]:
        """The selectable base accounts.

        :return: The selectable base accounts.
        """
        return BaseAccount.query\
            .filter(sa.func.char_length(BaseAccount.code) == 4)\
            .order_by(BaseAccount.code).all()


def sort_accounts_in(base_code: str, exclude: int) -> None:
    """Sorts the accounts under a base account after changing the base
    account or deleting an account.

    :param base_code: The code of the base account.
    :param exclude: The account ID to exclude.
    :return: None.
    """
    accounts: list[Account] = Account.query\
        .filter(Account.base_code == base_code,
                Account.id != exclude)\
        .order_by(Account.no).all()
    for i in range(len(accounts)):
        if accounts[i].no != i + 1:
            accounts[i].no = i + 1


class AccountReorderForm:
    """The form to reorder the accounts."""

    def __init__(self, base: BaseAccount):
        """Constructs the form to reorder the accounts under a base account.

        :param base: The base account.
        """
        self.base: BaseAccount = base
        """The base account."""
        self.is_modified: bool = False
        """Whether the order is modified."""

    def save_order(self) -> None:
        """Saves the order of the account.

        :return:
        """
        accounts: list[Account] = self.base.accounts

        # Collects the specified order.
        orders: dict[Account, int] = {}
        for account in accounts:
            if f"{account.id}-no" in request.form:
                try:
                    orders[account] = int(request.form[f"{account.id}-no"])
                except ValueError:
                    pass

        # Missing and invalid orders are appended to the end.
        missing: list[Account] = [x for x in accounts if x not in orders]
        if len(missing) > 0:
            next_no: int = 1 if len(orders) == 0 else max(orders.values()) + 1
            for account in missing:
                orders[account] = next_no

        # Sort by the specified order first, and their original order.
        accounts = sorted(accounts, key=lambda x: (orders[x], x.no, x.code))

        # Update the orders.
        with db.session.no_autoflush:
            for i in range(len(accounts)):
                if accounts[i].no != i + 1:
                    accounts[i].no = i + 1
                    self.is_modified = True



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/10

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The reorder forms for the journal entry management.

"""
import datetime as dt

import sqlalchemy as sa
from flask import request

from accounting import db
from accounting.models import JournalEntry


def sort_journal_entries_in(date: dt.date, exclude: int | None = None) -> None:
    """Sorts the journal entries under a date after changing the date or
    deleting a journal entry.

    :param date: The date of the journal entry.
    :param exclude: The journal entry ID to exclude.
    :return: None.
    """
    conditions: list[sa.BinaryExpression] = [JournalEntry.date == date]
    if exclude is not None:
        conditions.append(JournalEntry.id != exclude)
    journal_entries: list[JournalEntry] = JournalEntry.query\
        .filter(*conditions)\
        .order_by(JournalEntry.no).all()
    for i in range(len(journal_entries)):
        if journal_entries[i].no != i + 1:
            journal_entries[i].no = i + 1


class JournalEntryReorderForm:
    """The form to reorder the journal entries."""

    def __init__(self, date: dt.date):
        """Constructs the form to reorder the journal entries in a day.

        :param date: The date.
        """
        self.date: dt.date = date
        """The date."""
        self.is_modified: bool = False
        """Whether the order is modified."""

    def save_order(self) -> None:
        """Saves the order of the account.

        :return:
        """
        journal_entries: list[JournalEntry] = JournalEntry.query\
            .filter(JournalEntry.date == self.date).all()

        # Collects the specified order.
        orders: dict[JournalEntry, int] = {}
        for journal_entry in journal_entries:
            if f"{journal_entry.id}-no" in request.form:
                try:
                    orders[journal_entry] \
                        = int(request.form[f"{journal_entry.id}-no"])
                except ValueError:
                    pass

        # Missing and invalid orders are appended to the end.
        missing: list[JournalEntry] \
            = [x for x in journal_entries if x not in orders]
        if len(missing) > 0:
            next_no: int = 1 if len(orders) == 0 else max(orders.values()) + 1
            for journal_entry in missing:
                orders[journal_entry] = next_no

        # Sort by the specified order first, and their original order.
        journal_entries.sort(key=lambda x: (orders[x], x.no))

        # Update the orders.
        with db.session.no_autoflush:
            for i in range(len(journal_entries)):
                if journal_entries[i].no != i + 1:
                    journal_entries[i].no = i + 1
                    self.is_modified = True



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/4

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The named shortcut periods.

"""
import datetime as dt

from accounting.locale import gettext
from .month_end import month_end
from .period import Period


class ThisMonth(Period):
    """The period of this month."""
    def __init__(self):
        today: dt.date = dt.date.today()
        this_month_start: dt.date = dt.date(today.year, today.month, 1)
        super().__init__(this_month_start, month_end(today))
        self.is_default = True
        self.is_this_month = True

    def _set_properties(self) -> None:
        self.spec = "this-month"
        self.desc = gettext("This Month")
        self.is_a_month = True
        self.is_type_month = True


class LastMonth(Period):
    """The period of this month."""
    def __init__(self):
        today: dt.date = dt.date.today()
        year: int = today.year
        month: int = today.month - 1
        if month < 1:
            year = year - 1
            month = 12
        start: dt.date = dt.date(year, month, 1)
        super().__init__(start, month_end(start))
        self.is_last_month = True

    def _set_properties(self) -> None:
        self.spec = "last-month"
        self.desc = gettext("Last Month")
        self.is_a_month = True
        self.is_type_month = True


class SinceLastMonth(Period):
    """The period of this month."""
    def __init__(self):
        today: dt.date = dt.date.today()
        year: int = today.year
        month: int = today.month - 1
        if month < 1:
            year = year - 1
            month = 12
        start: dt.date = dt.date(year, month, 1)
        super().__init__(start, None)
        self.is_since_last_month = True

    def _set_properties(self) -> None:
        self.spec = "since-last-month"
        self.desc = gettext("Since Last Month")
        self.is_type_month = True


class ThisYear(Period):
    """The period of this year."""
    def __init__(self):
        year: int = dt.date.today().year
        start: dt.date = dt.date(year, 1, 1)
        end: dt.date = dt.date(year, 12, 31)
        super().__init__(start, end)
        self.is_this_year = True

    def _set_properties(self) -> None:
        self.spec = "this-year"
        self.desc = gettext("This Year")
        self.is_a_year = True


class LastYear(Period):
    """The period of last year."""
    def __init__(self):
        year: int = dt.date.today().year
        start: dt.date = dt.date(year - 1, 1, 1)
        end: dt.date = dt.date(year - 1, 12, 31)
        super().__init__(start, end)
        self.is_last_year = True

    def _set_properties(self) -> None:
        self.spec = "last-year"
        self.desc = gettext("Last Year")
        self.is_a_year = True


class Today(Period):
    """The period of today."""
    def __init__(self):
        today: dt.date = dt.date.today()
        super().__init__(today, today)
        self.is_today = True

    def _set_properties(self) -> None:
        self.spec = "today"
        self.desc = gettext("Today")
        self.is_a_day = True


class Yesterday(Period):
    """The period of yesterday."""
    def __init__(self):
        yesterday: dt.date = dt.date.today() - dt.timedelta(days=1)
        super().__init__(yesterday, yesterday)
        self.is_yesterday = True

    def _set_properties(self) -> None:
        self.spec = "yesterday"
        self.desc = gettext("Yesterday")
        self.is_a_day = True


class AllTime(Period):
    """The period of all time."""
    def __init__(self):
        super().__init__(None, None)
        self.is_all = True

    def _set_properties(self) -> None:
        self.spec = "all-time"
        self.desc = gettext("All")


class TemplatePeriod(Period):
    """The period template."""
    def __init__(self):
        super().__init__(None, None)

    def _set_properties(self) -> None:
        self.spec = "PERIOD"


class YearPeriod(Period):
    """A year period."""

    def __init__(self, year: int):
        """Constructs a year period.

        :param year: The year.
        """
        start: dt.date = dt.date(year, 1, 1)
        end: dt.date = dt.date(year, 12, 31)
        super().__init__(start, end)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/25

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The data models.

"""
from __future__ import annotations

import datetime as dt
import re
from decimal import Decimal
from typing import Type, Self

import sqlalchemy as sa
from babel import Locale
from flask_babel import get_locale, get_babel
from sqlalchemy import text
from sqlalchemy.orm import Mapped, mapped_column

from accounting import db
from accounting.locale import gettext
from accounting.utils.user import user_cls, user_pk_column


class BaseAccount(db.Model):
    """A base account."""
    __tablename__ = "accounting_base_accounts"
    """The table name."""
    code: Mapped[str] = mapped_column(primary_key=True)
    """The account code."""
    title_l10n: Mapped[str] = mapped_column("title")
    """The title."""
    l10n: Mapped[list[BaseAccountL10n]] \
        = db.relationship(back_populates="account", lazy=False)
    """The localized titles."""
    accounts: Mapped[list[Account]] = db.relationship(back_populates="base")
    """The descendant accounts under the base account."""

    def __str__(self) -> str:
        """Returns the string representation of the base account.

        :return: The string representation of the base account.
        """
        return f"{self.code} {self.title}"

    @property
    def title(self) -> str:
        """Returns the title in the current locale.

        :return: The title in the current locale.
        """
        current_locale: Locale = get_locale()
        if current_locale == get_babel().instance.default_locale:
            return self.title_l10n
        for l10n in self.l10n:
            if l10n.locale == str(current_locale):
                return l10n.title
        return self.title_l10n

    @property
    def query_values(self) -> list[str]:
        """Returns the values to be queried.

        :return: The values to be queried.
        """
        return [self.code, self.title_l10n] + [x.title for x in self.l10n]


class BaseAccountL10n(db.Model):
    """A localized base account title."""
    __tablename__ = "accounting_base_accounts_l10n"
    """The table name."""
    account_code: Mapped[str] \
        = mapped_column(db.ForeignKey(BaseAccount.code, onupdate="CASCADE",
                                      ondelete="CASCADE"),
                        primary_key=True)
    """The account code."""
    account: Mapped[BaseAccount] = db.relationship(back_populates="l10n")
    """The account."""
    locale: Mapped[str] = mapped_column(primary_key=True)
    """The locale."""
    title: Mapped[str]
    """The localized title."""


class Account(db.Model):
    """An account."""
    __tablename__ = "accounting_accounts"
    """The table name."""
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=False)
    """The account ID."""
    base_code: Mapped[str] \
        = mapped_column(db.ForeignKey(BaseAccount.code, onupdate="CASCADE",
                                      ondelete="CASCADE"))
    """The code of the base account."""
    base: Mapped[BaseAccount] = db.relationship(back_populates="accounts")
    """The base account."""
    no: Mapped[int] = mapped_column(default=text("1"))
    """The account number under the base account."""
    title_l10n: Mapped[str] = mapped_column("title")
    """The title."""
    is_need_offset: Mapped[bool] = mapped_column(default=False)
    """Whether the journal entry line items of this account need offset."""
    created_at: Mapped[dt.datetime] \
        = mapped_column(db.DateTime(timezone=True),
                        server_default=db.func.now())
    """The date and time when this record was created."""
    created_by_id: Mapped[int] \
        = mapped_column(db.ForeignKey(user_pk_column, onupdate="CASCADE"))
    """The ID of the user who created the record."""
    created_by: Mapped[user_cls] = db.relationship(foreign_keys=created_by_id)
    """The user who created the record."""
    updated_at: Mapped[dt.datetime] \
        = mapped_column(db.DateTime(timezone=True),
                        server_default=db.func.now())
    """The date and time when this record was last updated."""
    updated_by_id: Mapped[int] \
        = mapped_column(db.ForeignKey(user_pk_column, onupdate="CASCADE"))
    """The ID of the last user who updated the record."""
    updated_by: Mapped[user_cls] = db.relationship(foreign_keys=updated_by_id)
    """The last user who updated the record."""
    l10n: Mapped[list[AccountL10n]] \
        = db.relationship(back_populates="account", lazy=False)
    """The localized titles."""
    line_items: Mapped[list[JournalEntryLineItem]] \
        = db.relationship(back_populates="account")
    """The journal entry line items."""

    CASH_CODE: str = "1111-001"
    """The code of the cash account,"""
    ACCUMULATED_CHANGE_CODE: str = "3351-001"
    """The code of the accumulated-change account,"""
    NET_CHANGE_CODE: str = "3353-001"
    """The code of the net-change account,"""

    def __str__(self) -> str:
        """Returns the string representation of this account.

        :return: The string representation of this account.
        """
        return f"{self.base_code}-{self.no:03d} {self.title}"

    @property
    def code(self) -> str:
        """Returns the code.

        :return: The code.
        """
        return f"{self.base_code}-{self.no:03d}"

    @property
    def title(self) -> str:
        """Returns the title in the current locale.

        :return: The title in the current locale.
        """
        current_locale: Locale = get_locale()
        if current_locale == get_babel().instance.default_locale:
            return self.title_l10n
        for l10n in self.l10n:
            if l10n.locale == str(current_locale):
                return l10n.title
        return self.title_l10n

    @title.setter
    def title(self, value: str) -> None:
        """Sets the title in the current locale.

        :param value: The new title.
        :return: None.
        """
        if self.title == value:
            return
        if self.title_l10n is None:
            self.title_l10n = value
            return
        current_locale: Locale = get_locale()
        if current_locale == get_babel().instance.default_locale:
            self.title_l10n = value
            return
        for l10n in self.l10n:
            if l10n.locale == str(current_locale):
                l10n.title = value
                return
        self.l10n.append(AccountL10n(locale=str(current_locale), title=value))

    @property
    def is_real(self) -> bool:
        """Returns whether the account is a real account.

        :return: True if the account is a real account, or False otherwise.
        """
        return self.base_code[0] in {"1", "2", "3"}

    @property
    def is_nominal(self) -> bool:
        """Returns whether the account is a nominal account.

        :return: True if the account is a nominal account, or False otherwise.
        """
        return not self.is_real

    @property
    def count(self) -> int:
        """Returns the number of items in the account.

        :return: The number of items in the account.
        """
        if not hasattr(self, "__count"):
            setattr(self, "__count", 0)
        return getattr(self, "__count")

    @count.setter
    def count(self, value: int) -> None:
        """Sets the number of items in the account.

        :param value: The number of items in the account.
        :return: None.
        """
        setattr(self, "__count", value)

    @property
    def query_values(self) -> list[str]:
        """Returns the values to be queried.

        :return: The values to be queried.
        """
        return [self.code, self.title_l10n] + [x.title for x in self.l10n]

    @property
    def is_modified(self) -> bool:
        """Returns whether a product account was modified.

        :return: True if modified, or False otherwise.
        """
        if db.session.is_modified(self):
            return True
        for l10n in self.l10n:
            if db.session.is_modified(l10n):
                return True
        return False

    @property
    def can_delete(self) -> bool:
        """Returns whether the account can be deleted.

        :return: True if the account can be deleted, or False otherwise.
        """
        if self.code in {"1111-001", "3351-001", "3353-001"}:
            return False
        return len(self.line_items) == 0

    def delete(self) -> None:
        """Deletes this account.

        :return: None.
        """
        AccountL10n.query.filter(AccountL10n.account == self).delete()
        cls: Type[Self] = self.__class__
        cls.query.filter(cls.id == self.id).delete()

    @classmethod
    def find_by_code(cls, code: str) -> Self | None:
        """Finds an account by its code.

        :param code: The code.
        :return: The account, or None if this account does not exist.
        """
        m = re.match(r"^([1-9]{4})-(\d{3})$", code)
        if m is None:
            return None
        return cls.query.filter(cls.base_code == m.group(1),
                                cls.no == int(m.group(2))).first()

    @classmethod
    def selectable_debit(cls) -> list[Self]:
        """Returns the selectable debit accounts.
        Payable line items can not start from debit.

        :return: The selectable debit accounts.
        """
        return cls.query.filter(sa.or_(cls.base_code.startswith("1"),
                                       sa.and_(cls.base_code.startswith("2"),
                                               sa.not_(cls.is_need_offset)),
                                       cls.base_code.startswith("3"),
                                       cls.base_code.startswith("5"),
                                       cls.base_code.startswith("6"),
                                       cls.base_code.startswith("75"),
                                       cls.base_code.startswith("76"),
                                       cls.base_code.startswith("77"),
                                       cls.base_code.startswith("78"),
                                       cls.base_code.startswith("8"),
                                       cls.base_code.startswith("9")),
                                cls.base_code != "3351",
                                cls.base_code != "3353")\
            .order_by(cls.base_code, cls.no).all()

    @classmethod
    def selectable_credit(cls) -> list[Self]:
        """Returns the selectable debit accounts.
        Receivable line items can not start from credit.

        :return: The selectable debit accounts.
        """
        return cls.query.filter(sa.or_(sa.and_(cls.base_code.startswith("1"),
                                               sa.not_(cls.is_need_offset)),
                                       cls.base_code.startswith("2"),
                                       cls.base_code.startswith("3"),
                                       cls.base_code.startswith("4"),
                                       cls.base_code.startswith("71"),
                                       cls.base_code.startswith("72"),
                                       cls.base_code.startswith("73"),
                                       cls.base_code.startswith("74"),
                                       cls.base_code.startswith("8"),
                                       cls.base_code.startswith("9")),
                                cls.base_code != "3351",
                                cls.base_code != "3353")\
            .order_by(cls.base_code, cls.no).all()

    @classmethod
    def cash(cls) -> Self:
        """Returns the cash account.

        :return: The cash account
        """
        return cls.find_by_code(cls.CASH_CODE)

    @classmethod
    def accumulated_change(cls) -> Self:
        """Returns the accumulated-change account.

        :return: The accumulated-change account
        """
        return cls.find_by_code(cls.ACCUMULATED_CHANGE_CODE)


class AccountL10n(db.Model):
    """A localized account title."""
    __tablename__ = "accounting_accounts_l10n"
    """The table name."""
    account_id: Mapped[int] \
        = mapped_column(db.ForeignKey(Account.id, onupdate="CASCADE",
                                      ondelete="CASCADE"),
                        primary_key=True)
    """The account ID."""
    account: Mapped[Account] = db.relationship(back_populates="l10n")
    """The account."""
    locale: Mapped[str] = mapped_column(primary_key=True)
    """The locale."""
    title: Mapped[str]
    """The localized title."""


class Currency(db.Model):
    """A currency."""
    __tablename__ = "accounting_currencies"
    """The table name."""
    code: Mapped[str] = mapped_column(primary_key=True)
    """The currency code."""
    name_l10n: Mapped[str] = mapped_column("name")
    """The currency name."""
    created_at: Mapped[dt.datetime] \
        = mapped_column(db.DateTime(timezone=True),
                        server_default=db.func.now())
    """The date and time when this record was created."""
    created_by_id: Mapped[int] \
        = mapped_column(db.ForeignKey(user_pk_column, onupdate="CASCADE"))
    """The ID of the user who created the record."""
    created_by: Mapped[user_cls] = db.relationship(foreign_keys=created_by_id)
    """The user who created the record."""
    updated_at: Mapped[dt.datetime] \
        = mapped_column(db.DateTime(timezone=True),
                        server_default=db.func.now())
    """The date and time when this record was last updated."""
    updated_by_id: Mapped[int] \
        = mapped_column(db.ForeignKey(user_pk_column, onupdate="CASCADE"))
    """The ID of the last user who updated the record."""
    updated_by: Mapped[user_cls] \
        = db.relationship(foreign_keys=updated_by_id)
    """The last user who updated the record."""
    l10n: Mapped[list[CurrencyL10n]] \
        = db.relationship(back_populates="currency", lazy=False)
    """The localized names."""
    line_items: Mapped[list[JournalEntryLineItem]] \
        = db.relationship(back_populates="currency")
    """The journal entry line items."""

    def __str__(self) -> str:
        """Returns the string representation of the currency.

        :return: The string representation of the currency.
        """
        return f"{self.name.title()} ({self.code})"

    @property
    def name(self) -> str:
        """Returns the name in the current locale.

        :return: The name in the current locale.
        """
        current_locale: Locale = get_locale()
        if current_locale == get_babel().instance.default_locale:
            return self.name_l10n
        for l10n in self.l10n:
            if l10n.locale == str(current_locale):
                return l10n.name
        return self.name_l10n

    @name.setter
    def name(self, value: str) -> None:
        """Sets the name in the current locale.

        :param value: The new name.
        :return: None.
        """
        if self.name == value:
            return
        if self.name_l10n is None:
            self.name_l10n = value
            return
        current_locale: Locale = get_locale()
        if current_locale == get_babel().instance.default_locale:
            self.name_l10n = value
            return
        for l10n in self.l10n:
            if l10n.locale == str(current_locale):
                l10n.name = value
                return
        self.l10n.append(CurrencyL10n(locale=str(current_locale), name=value))

    @property
    def is_modified(self) -> bool:
        """Returns whether a product account was modified.

        :return: True if modified, or False otherwise.
        """
        if db.session.is_modified(self):
            return True
        for l10n in self.l10n:
            if db.session.is_modified(l10n):
                return True
        return False

    @property
    def can_delete(self) -> bool:
        """Returns whether the currency can be deleted.

        :return: True if the currency can be deleted, or False otherwise.
        """
        from accounting.template_globals import default_currency_code
        if self.code == default_currency_code():
            return False
        return len(self.line_items) == 0

    def delete(self) -> None:
        """Deletes the currency.

        :return: None.
        """
        CurrencyL10n.query.filter(CurrencyL10n.currency == self).delete()
        cls: Type[Self] = self.__class__
        cls.query.filter(cls.code == self.code).delete()


class CurrencyL10n(db.Model):
    """A localized currency name."""
    __tablename__ = "accounting_currencies_l10n"
    """The table name."""
    currency_code: Mapped[str] \
        = mapped_column(db.ForeignKey(Currency.code, onupdate="CASCADE",
                                      ondelete="CASCADE"),
                        primary_key=True)
    """The currency code."""
    currency: Mapped[Currency] = db.relationship(back_populates="l10n")
    """The currency."""
    locale: Mapped[str] = mapped_column(primary_key=True)
    """The locale."""
    name: Mapped[str]
    """The localized name."""


class JournalEntryCurrency:
    """A currency in a journal entry."""

    def __init__(self, code: str, debit: list[JournalEntryLineItem],
                 credit: list[JournalEntryLineItem]):
        """Constructs the currency in the journal entry.

        :param code: The currency code.
        :param debit: The debit line items.
        :param credit: The credit line items.
        """
        self.code: str = code
        """The currency code."""
        self.debit: list[JournalEntryLineItem] = debit
        """The debit line items."""
        self.credit: list[JournalEntryLineItem] = credit
        """The credit line items."""

    @property
    def name(self) -> str:
        """Returns the currency name.

        :return: The currency name.
        """
        return db.session.get(Currency, self.code).name

    @property
    def debit_total(self) -> Decimal:
        """Returns the total amount of the debit line items.

        :return: The total amount of the debit line items.
        """
        return sum([x.amount for x in self.debit])

    @property
    def credit_total(self) -> str:
        """Returns the total amount of the credit line items.

        :return: The total amount of the credit line items.
        """
        return sum([x.amount for x in self.credit])


class JournalEntry(db.Model):
    """A journal entry."""
    __tablename__ = "accounting_journal_entries"
    """The table name."""
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=False)
    """The journal entry ID."""
    date: Mapped[dt.date]
    """The date."""
    no: Mapped[int] = mapped_column(default=text("1"))
    """The journal entry number under the date."""
    note: Mapped[str | None]
    """The note."""
    created_at: Mapped[dt.datetime] \
        = mapped_column(db.DateTime(timezone=True),
                        server_default=db.func.now())
    """The date and time when this record was created."""
    created_by_id: Mapped[int] \
        = mapped_column(db.ForeignKey(user_pk_column, onupdate="CASCADE"))
    """The ID of the user who created the record."""
    created_by: Mapped[user_cls] = db.relationship(foreign_keys=created_by_id)
    """The user who created the record."""
    updated_at: Mapped[dt.datetime] \
        = mapped_column(db.DateTime(timezone=True),
                        server_default=db.func.now())
    """The date and time when this record was last updated."""
    updated_by_id: Mapped[int] \
        = mapped_column(db.ForeignKey(user_pk_column, onupdate="CASCADE"))
    """The ID of the last user who updated the record."""
    updated_by: Mapped[user_cls] = db.relationship(foreign_keys=updated_by_id)
    """The last user who updated the record."""
    line_items: Mapped[list[JournalEntryLineItem]] \
        = db.relationship(back_populates="journal_entry")
    """The line items."""

    def __str__(self) -> str:
        """Returns the string representation of this journal entry.

        :return: The string representation of this journal entry.
        """
        if self.is_cash_disbursement:
            return gettext("Cash Disbursement Journal Entry#%(id)s",
                           id=self.id)
        if self.is_cash_receipt:
            return gettext("Cash Receipt Journal Entry#%(id)s", id=self.id)
        return gettext("Transfer Journal Entry#%(id)s", id=self.id)

    @property
    def currencies(self) -> list[JournalEntryCurrency]:
        """Returns the line items categorized by their currencies.

        :return: The currency categories.
        """
        line_items: list[JournalEntryLineItem] = sorted(self.line_items,
                                                        key=lambda x: x.no)
        codes: list[str] = []
        by_currency: dict[str, list[JournalEntryLineItem]] = {}
        for line_item in line_items:
            if line_item.currency_code not in by_currency:
                codes.append(line_item.currency_code)
                by_currency[line_item.currency_code] = []
            by_currency[line_item.currency_code].append(line_item)
        return [JournalEntryCurrency(code=x,
                                     debit=[y for y in by_currency[x]
                                            if y.is_debit],
                                     credit=[y for y in by_currency[x]
                                             if not y.is_debit])
                for x in codes]

    @property
    def is_cash_receipt(self) -> bool:
        """Returns whether this is a cash receipt journal entry.

        :return: True if this is a cash receipt journal entry, or False
            otherwise.
        """
        for currency in self.currencies:
            if len(currency.debit) > 1:
                return False
            if currency.debit[0].account.code != Account.CASH_CODE:
                return False
        return True

    @property
    def is_cash_disbursement(self) -> bool:
        """Returns whether this is a cash disbursement journal entry.

        :return: True if this is a cash disbursement journal entry, or False
            otherwise.
        """
        for currency in self.currencies:
            if len(currency.credit) > 1:
                return False
            if currency.credit[0].account.code != Account.CASH_CODE:
                return False
        return True

    @property
    def can_delete(self) -> bool:
        """Returns whether the journal entry can be deleted.

        :return: True if the journal entry can be deleted, or False otherwise.
        """
        for line_item in self.line_items:
            if len(line_item.offsets) > 0:
                return False
        return True

    def delete(self) -> None:
        """Deletes the journal entry.

        :return: None.
        """
        JournalEntryLineItem.query\
            .filter(JournalEntryLineItem.journal_entry_id == self.id).delete()
        db.session.delete(self)


class JournalEntryLineItem(db.Model):
    """A line item in the journal entry."""
    __tablename__ = "accounting_journal_entry_line_items"
    """The table name."""
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=False)
    """The line item ID."""
    journal_entry_id: Mapped[int] \
        = mapped_column(db.ForeignKey(JournalEntry.id, onupdate="CASCADE",
                                      ondelete="CASCADE"))
    """The journal entry ID."""
    journal_entry: Mapped[JournalEntry] \
        = db.relationship(back_populates="line_items")
    """The journal entry."""
    is_debit: Mapped[bool]
    """True for a debit line item, or False for a credit line item."""
    no: Mapped[int]
    """The line item number under the journal entry and debit or credit."""
    original_line_item_id: Mapped[int | None] \
        = mapped_column(db.ForeignKey(id, onupdate="CASCADE"))
    """The ID of the original line item."""
    original_line_item: Mapped[JournalEntryLineItem | None] \
        = db.relationship(remote_side=id, passive_deletes=True)
    """The original line item."""
    currency_code: Mapped[str] \
        = mapped_column(db.ForeignKey(Currency.code, onupdate="CASCADE"))
    """The currency code."""
    currency: Mapped[Currency] = db.relationship(back_populates="line_items")
    """The currency."""
    account_id: Mapped[int] \
        = mapped_column(db.ForeignKey(Account.id, onupdate="CASCADE"))
    """The account ID."""
    account: Mapped[Account] \
        = db.relationship(back_populates="line_items", lazy=False)
    """The account."""
    description: Mapped[str | None]
    """The description."""
    amount: Mapped[Decimal] = mapped_column(db.Numeric(14, 2))
    """The amount."""

    def __str__(self) -> str:
        """Returns the string representation of the line item.

        :return: The string representation of the line item.
        """
        if not hasattr(self, "__str"):
            from accounting.template_filters import format_date, format_amount
            setattr(self, "__str",
                    gettext("%(date)s %(description)s %(amount)s",
                            date=format_date(self.journal_entry.date),
                            description="" if self.description is None
                            else self.description,
                            amount=format_amount(self.amount)))
        return getattr(self, "__str")

    @property
    def account_code(self) -> str:
        """Returns the account code.

        :return: The account code.
        """
        return self.account.code

    @property
    def is_need_offset(self) -> bool:
        """Returns whether the line item needs offset.

        :return: True if the line item needs offset, or False otherwise.
        """
        if not self.account.is_need_offset:
            return False
        if self.account.base_code[0] == "1" and not self.is_debit:
            return False
        if self.account.base_code[0] == "2" and self.is_debit:
            return False
        return True

    @property
    def debit(self) -> Decimal | None:
        """Returns the debit amount.

        :return: The debit amount, or None if this is not a debit line item.
        """
        if not hasattr(self, "__debit"):
            setattr(self, "__debit", self.amount if self.is_debit else None)
        return getattr(self, "__debit")

    @debit.setter
    def debit(self, value: Decimal | None) -> None:
        """Sets the debit amount.

        :param value: The debit amount.
        :return: None.
        """
        setattr(self, "__debit", value)

    @property
    def credit(self) -> Decimal | None:
        """Returns the credit amount.

        :return: The credit amount, or None if this is not a credit line item.
        """
        if not hasattr(self, "__credit"):
            setattr(self, "__credit", None if self.is_debit else self.amount)
        return getattr(self, "__credit")

    @credit.setter
    def credit(self, value: Decimal | None) -> None:
        """Sets the credit amount.

        :param value: The credit amount.
        :return: None.
        """
        setattr(self, "__credit", value)

    @property
    def net_balance(self) -> Decimal:
        """Returns the net balance.

        :return: The net balance.
        """
        if not hasattr(self, "__net_balance"):
            setattr(self, "__net_balance", self.amount + sum(
                [x.amount if x.is_debit == self.is_debit else -x.amount
                 for x in self.offsets]))
        return getattr(self, "__net_balance")

    @net_balance.setter
    def net_balance(self, value: Decimal) -> None:
        """Sets the net balance.

        :param value: The net balance.
        :return: None.
        """
        setattr(self, "__net_balance", value)

    @property
    def balance(self) -> Decimal:
        """Returns the balance.

        :return: The balance.
        """
        if not hasattr(self, "__balance"):
            setattr(self, "__balance", Decimal("0"))
        return getattr(self, "__balance")

    @balance.setter
    def balance(self, value: Decimal) -> None:
        """Sets the balance.

        :param value: The balance.
        :return: None.
        """
        setattr(self, "__balance", value)

    @property
    def offsets(self) -> list[Self]:
        """Returns the offset items.

        :return: The offset items.
        """
        if not hasattr(self, "__offsets"):
            cls: Type[Self] = self.__class__
            offsets: list[Self] = cls.query.join(JournalEntry)\
                .filter(JournalEntryLineItem.original_line_item_id == self.id)\
                .order_by(JournalEntry.date, JournalEntry.no,
                          cls.is_debit, cls.no).all()
            setattr(self, "__offsets", offsets)
        return getattr(self, "__offsets")

    @property
    def is_offset(self) -> bool:
        """Returns whether the line item is an offset.

        :return: True if the line item is an offset, or False otherwise.
        """
        if not hasattr(self, "__is_offset"):
            setattr(self, "__is_offset", False)
        return getattr(self, "__is_offset")

    @is_offset.setter
    def is_offset(self, value: bool) -> None:
        """Sets whether the line item is an offset.

        :param value: True if the line item is an offset, or False otherwise.
        :return: None.
        """
        setattr(self, "__is_offset", value)

    @property
    def match(self) -> Self | None:
        """Returns the match of the line item.

        :return: The match of the line item.
        """
        if not hasattr(self, "__match"):
            setattr(self, "__match", None)
        return getattr(self, "__match")

    @match.setter
    def match(self, value: Self) -> None:
        """Sets the match of the line item.

        :param value: The matcho of the line item.
        :return: None.
        """
        setattr(self, "__match", value)

    @property
    def query_values(self) -> list[str]:
        """Returns the values to be queried.

        :return: The values to be queried.
        """
        def format_amount(value: Decimal) -> str:
            whole: int = int(value)
            frac: Decimal = (value - whole).normalize()
            return str(whole) + str(abs(frac))[1:]

        return ["{}/{}/{}".format(self.journal_entry.date.year,
                                  self.journal_entry.date.month,
                                  self.journal_entry.date.day),
                "" if self.description is None else self.description,
                str(self.account),
                format_amount(self.amount)]


class Option(db.Model):
    """An option."""
    __tablename__ = "accounting_options"
    """The table name."""
    name: Mapped[str] = mapped_column(primary_key=True)
    """The name."""
    value: Mapped[str] = mapped_column(db.Text)
    """The option value."""
    created_at: Mapped[dt.datetime] \
        = mapped_column(db.DateTime(timezone=True),
                        server_default=db.func.now())
    """The date and time when this record was created."""
    created_by_id: Mapped[int] \
        = mapped_column(db.ForeignKey(user_pk_column, onupdate="CASCADE"))
    """The ID of the user who created the record."""
    created_by: Mapped[user_cls] = db.relationship(foreign_keys=created_by_id)
    """The user who created the record."""
    updated_at: Mapped[dt.datetime] \
        = mapped_column(db.DateTime(timezone=True),
                        server_default=db.func.now())
    """The date and time when this record was last updated."""
    updated_by_id: Mapped[int] \
        = mapped_column(db.ForeignKey(user_pk_column, onupdate="CASCADE"))
    """The ID of the last user who updated the record."""
    updated_by: Mapped[user_cls] = db.relationship(foreign_keys=updated_by_id)
    """The last user who updated the record."""



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/25

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The permissions.

This module should not import any other module from the application.

"""
from collections.abc import Callable

from flask import abort, Blueprint, Response

from accounting.utils.user import get_current_user, UserUtilityInterface


def has_permission(rule: Callable[[], bool]) -> Callable:
    """The permission decorator to check whether the current user is allowed.

    :param rule: The permission rule.
    :return: The view decorator.
    """

    def decorator(view: Callable) -> Callable:
        """The view decorator to decorate a view with permission tests.

        :param view: The view.
        :return: The decorated view.
        """

        def decorated_view(*args, **kwargs):
            """The decorated view that tests against a permission rule.

            :param args: The arguments of the view.
            :param kwargs: The keyword arguments of the view.
            :return: The response of the view.
            :raise Forbidden: When the user is denied.
            """
            if not rule():
                if get_current_user() is None:
                    response: Response | None = _unauthorized_func()
                    if response is not None:
                        return response
                abort(403)
            return view(*args, **kwargs)

        return decorated_view

    return decorator


__can_view_func: Callable[[], bool] = lambda: True
"""The callback that returns whether the current user can view the accounting
data."""
__can_edit_func: Callable[[], bool] = lambda: True
"""The callback that returns whether the current user can edit the accounting
data."""
__can_admin_func: Callable[[], bool] = lambda: True
"""The callback that returns whether the current user can administrate the
accounting settings."""
_unauthorized_func: Callable[[], Response | None] \
    = lambda: Response(status=403)
"""The callback that returns the response to require the user to log in."""


def can_view() -> bool:
    """Returns whether the current user can view the accounting data.

    :return: True if the current user can view the accounting data, or False
        otherwise.
    """
    return __can_view_func()


def can_edit() -> bool:
    """Returns whether the current user can edit the accounting data.

    The user has to log in.

    :return: True if the current user can edit the accounting data, or False
        otherwise.
    """
    if get_current_user() is None:
        return False
    return __can_edit_func()


def can_admin() -> bool:
    """Returns whether the current user can administrate the accounting
    settings.

    The user has to log in.

    :return: True if the current user can administrate the accounting settings,
        or False otherwise.
    """
    if get_current_user() is None:
        return False
    return __can_admin_func()


def init_app(bp: Blueprint, user_utils: UserUtilityInterface) -> None:
    """Initializes the application.

    :param bp: The blueprint of the accounting application.
    :param user_utils: The user utilities.
    :return: None.
    """
    global __can_view_func, __can_edit_func, __can_admin_func, \
        _unauthorized_func
    __can_view_func = user_utils.can_view
    __can_edit_func = user_utils.can_edit
    __can_admin_func = user_utils.can_admin
    _unauthorized_func = user_utils.unauthorized
    bp.add_app_template_global(user_utils.can_view, "accounting_can_view")
    bp.add_app_template_global(user_utils.can_edit, "accounting_can_edit")
    bp.add_app_template_global(user_utils.can_admin, "accounting_can_admin")



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/18

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The views for the journal entry management.

"""
import datetime as dt
from urllib.parse import parse_qsl, urlencode

import sqlalchemy as sa
from flask import Blueprint, render_template, session, redirect, request, \
    flash, url_for
from werkzeug.datastructures import ImmutableMultiDict

from accounting import db
from accounting.locale import lazy_gettext
from accounting.models import JournalEntry
from accounting.utils.cast import s
from accounting.utils.flash_errors import flash_form_errors
from accounting.utils.journal_entry_types import JournalEntryType
from accounting.utils.next_uri import inherit_next, or_next
from accounting.utils.permission import has_permission, can_view, can_edit
from accounting.utils.user import get_current_user_pk
from .forms import sort_journal_entries_in, JournalEntryReorderForm
from .template_filters import with_type, to_transfer, format_amount_input, \
    text2html
from .utils.operators import JournalEntryOperator, JOURNAL_ENTRY_TYPE_TO_OP, \
    get_journal_entry_op

bp: Blueprint = Blueprint("journal-entry", __name__)
"""The view blueprint for the journal entry management."""
bp.add_app_template_filter(with_type, "accounting_journal_entry_with_type")
bp.add_app_template_filter(to_transfer, "accounting_journal_entry_to_transfer")
bp.add_app_template_filter(format_amount_input,
                           "accounting_journal_entry_format_amount_input")
bp.add_app_template_filter(text2html, "accounting_journal_entry_text2html")


@bp.get("create/<journalEntryType:journal_entry_type>", endpoint="create")
@has_permission(can_edit)
def show_add_journal_entry_form(journal_entry_type: JournalEntryType) -> str:
    """Shows the form to add a journal entry.

    :param journal_entry_type: The journal entry type.
    :return: The form to add a journal entry.
    """
    journal_entry_op: JournalEntryOperator \
        = JOURNAL_ENTRY_TYPE_TO_OP[journal_entry_type]
    form: journal_entry_op.form
    if "form" in session:
        form = journal_entry_op.form(
            ImmutableMultiDict(parse_qsl(session["form"])))
        del session["form"]
        form.validate()
    else:
        form = journal_entry_op.form()
        form.date.data = dt.date.today()
    return journal_entry_op.render_create_template(form)


@bp.post("store/<journalEntryType:journal_entry_type>", endpoint="store")
@has_permission(can_edit)
def add_journal_entry(journal_entry_type: JournalEntryType) -> redirect:
    """Adds a journal entry.

    :param journal_entry_type: The journal entry type.
    :return: The redirection to the journal entry detail on success, or the
        journal entry creation form on error.
    """
    journal_entry_op: JournalEntryOperator \
        = JOURNAL_ENTRY_TYPE_TO_OP[journal_entry_type]
    form: journal_entry_op.form = journal_entry_op.form(request.form)
    if not form.validate():
        flash_form_errors(form)
        session["form"] = urlencode(list(request.form.items()))
        return redirect(inherit_next(with_type(
            url_for("accounting.journal-entry.create",
                    journal_entry_type=journal_entry_type))))
    journal_entry: JournalEntry = JournalEntry()
    form.populate_obj(journal_entry)
    db.session.add(journal_entry)
    db.session.commit()
    flash(s(lazy_gettext("The journal entry is added successfully.")),
          "success")
    return redirect(inherit_next(__get_detail_uri(journal_entry)))


@bp.get("<journalEntry:journal_entry>", endpoint="detail")
@has_permission(can_view)
def show_journal_entry_detail(journal_entry: JournalEntry) -> str:
    """Shows the journal entry detail.

    :param journal_entry: The journal entry.
    :return: The detail.
    """
    journal_entry_op: JournalEntryOperator \
        = get_journal_entry_op(journal_entry)
    return journal_entry_op.render_detail_template(journal_entry)


@bp.get("<journalEntry:journal_entry>/edit", endpoint="edit")
@has_permission(can_edit)
def show_journal_entry_edit_form(journal_entry: JournalEntry) -> str:
    """Shows the form to edit a journal entry.

    :param journal_entry: The journal entry.
    :return: The form to edit the journal entry.
    """
    journal_entry_op: JournalEntryOperator \
        = get_journal_entry_op(journal_entry, is_check_as=True)
    form: journal_entry_op.form
    if "form" in session:
        form = journal_entry_op.form(
            ImmutableMultiDict(parse_qsl(session["form"])))
        del session["form"]
        form.obj = journal_entry
        form.validate()
    else:
        form = journal_entry_op.form(obj=journal_entry)
    return journal_entry_op.render_edit_template(journal_entry, form)


@bp.post("<journalEntry:journal_entry>/update", endpoint="update")
@has_permission(can_edit)
def update_journal_entry(journal_entry: JournalEntry) -> redirect:
    """Updates a journal entry.

    :param journal_entry: The journal entry.
    :return: The redirection to the journal entry detail on success, or the
        journal entry edit form on error.
    """
    journal_entry_op: JournalEntryOperator \
        = get_journal_entry_op(journal_entry, is_check_as=True)
    form: journal_entry_op.form = journal_entry_op.form(request.form)
    form.obj = journal_entry
    if not form.validate():
        flash_form_errors(form)
        session["form"] = urlencode(list(request.form.items()))
        return redirect(inherit_next(with_type(
            url_for("accounting.journal-entry.edit",
                    journal_entry=journal_entry))))
    with db.session.no_autoflush:
        form.populate_obj(journal_entry)
    if not form.is_modified:
        flash(s(lazy_gettext("The journal entry was not modified.")),
              "success")
        return redirect(inherit_next(__get_detail_uri(journal_entry)))
    journal_entry.updated_by_id = get_current_user_pk()
    journal_entry.updated_at = sa.func.now()
    db.session.commit()
    flash(s(lazy_gettext("The journal entry is updated successfully.")),
          "success")
    return redirect(inherit_next(__get_detail_uri(journal_entry)))


@bp.post("<journalEntry:journal_entry>/delete", endpoint="delete")
@has_permission(can_edit)
def delete_journal_entry(journal_entry: JournalEntry) -> redirect:
    """Deletes a journal entry.

    :param journal_entry: The journal entry.
    :return: The redirection to the journal entry list on success, or the
        journal entry detail on error.
    """
    if not journal_entry.can_delete:
        flash(s(lazy_gettext("The journal entry cannot be deleted.")), "error")
        return redirect(inherit_next(__get_detail_uri(journal_entry)))
    journal_entry.delete()
    sort_journal_entries_in(journal_entry.date, journal_entry.id)
    db.session.commit()
    flash(s(lazy_gettext("The journal entry is deleted successfully.")),
          "success")
    return redirect(or_next(__get_default_page_uri()))


@bp.get("dates/<date:date>", endpoint="order")
@has_permission(can_view)
def show_journal_entry_order(date: dt.date) -> str:
    """Shows the order of the journal entries in a same date.

    :param date: The date.
    :return: The order of the journal entries in the date.
    """
    journal_entries: list[JournalEntry] = JournalEntry.query \
        .filter(JournalEntry.date == date) \
        .order_by(JournalEntry.no).all()
    return render_template("accounting/journal-entry/order.html",
                           date=date, list=journal_entries)


@bp.post("dates/<date:date>", endpoint="sort")
@has_permission(can_edit)
def sort_journal_entries(date: dt.date) -> redirect:
    """Reorders the journal entries in a date.

    :param date: The date.
    :return: The redirection to the incoming account or the account list.  The
        reordering operation does not fail.
    """
    form: JournalEntryReorderForm = JournalEntryReorderForm(date)
    form.save_order()
    if not form.is_modified:
        flash(s(lazy_gettext("The order was not modified.")), "success")
        return redirect(or_next(__get_default_page_uri()))
    db.session.commit()
    flash(s(lazy_gettext("The order is updated successfully.")), "success")
    return redirect(or_next(__get_default_page_uri()))


def __get_detail_uri(journal_entry: JournalEntry) -> str:
    """Returns the detail URI of a journal entry.

    :param journal_entry: The journal entry.
    :return: The detail URI of the journal entry.
    """
    return url_for("accounting.journal-entry.detail",
                   journal_entry=journal_entry)


def __get_default_page_uri() -> str:
    """Returns the URI for the default page.

    :return: The URI for the default page.
    """
    return url_for("accounting-report.default")
